# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import threading
import os
import sys
import re
import shutil
import unicodedata

from time import sleep
from subprocess import call, PIPE

from powerline.lib import add_divider_highlight_group
from powerline.lib.dict import mergedicts, REMOVE_THIS_KEY
from powerline.lib.humanize_bytes import humanize_bytes
from powerline.lib.vcs import guess, get_fallback_create_watcher
from powerline.lib.threaded import ThreadedSegment, KwThreadedSegment
from powerline.lib.monotonic import monotonic
from powerline.lib.vcs.git import git_directory
from powerline.lib.shell import run_cmd

import powerline.lib.unicode as plu

from tests.modules.lib import Pl, replace_attr
from tests.modules import TestCase, SkipTest


try:
	__import__('bzrlib')
except ImportError:
	use_bzr = False
else:
	use_bzr = True

try:
	__import__('hglib')
except ImportError:
	use_mercurial = False
else:
	use_mercurial = True


GIT_REPO = 'git_repo'
HG_REPO = 'hg_repo'
BZR_REPO = 'bzr_repo'


def thread_number():
	return len(threading.enumerate())


class TestShell(TestCase):
	def test_run_cmd(self):
		pl = Pl()
		self.assertEqual(run_cmd(pl, ['xxx_nonexistent_command_xxx']), None)
		self.assertEqual(len(pl.exceptions), 1)
		pl = Pl()
		self.assertEqual(run_cmd(pl, ['echo', '  test  ']), 'test')
		self.assertFalse(pl)
		self.assertEqual(run_cmd(pl, ['echo', '  test  '], strip=True), 'test')
		self.assertFalse(pl)
		self.assertEqual(run_cmd(pl, ['echo', '  test  '], strip=False), '  test  \n')
		self.assertFalse(pl)
		self.assertEqual(run_cmd(pl, ['cat'], stdin='test'), 'test')
		self.assertFalse(pl)
		self.assertEqual(run_cmd(pl, ['sh', '-c', 'cat >&2'], stdin='test'), '')
		self.assertFalse(pl)


class TestThreaded(TestCase):
	def test_threaded_segment(self):
		log = []
		pl = Pl()
		updates = [(None,)]
		lock = threading.Lock()
		event = threading.Event()
		block_event = threading.Event()

		class TestSegment(ThreadedSegment):
			interval = 10

			def set_state(self, **kwargs):
				event.clear()
				log.append(('set_state', kwargs))
				return super(TestSegment, self).set_state(**kwargs)

			def update(self, update_value):
				block_event.wait()
				event.set()
				# Make sleep first to prevent some race conditions
				log.append(('update', update_value))
				with lock:
					ret = updates[0]
				if isinstance(ret, Exception):
					raise ret
				else:
					return ret[0]

			def render(self, update, **kwargs):
				log.append(('render', update, kwargs))
				if isinstance(update, Exception):
					raise update
				else:
					return update

		# Non-threaded tests
		segment = TestSegment()
		block_event.set()
		updates[0] = (None,)
		self.assertEqual(segment(pl=pl), None)
		self.assertEqual(thread_number(), 1)
		self.assertEqual(log, [
			('set_state', {}),
			('update', None),
			('render', None, {'pl': pl, 'update_first': True}),
		])
		log[:] = ()

		segment = TestSegment()
		block_event.set()
		updates[0] = ('abc',)
		self.assertEqual(segment(pl=pl), 'abc')
		self.assertEqual(thread_number(), 1)
		self.assertEqual(log, [
			('set_state', {}),
			('update', None),
			('render', 'abc', {'pl': pl, 'update_first': True}),
		])
		log[:] = ()

		segment = TestSegment()
		block_event.set()
		updates[0] = ('abc',)
		self.assertEqual(segment(pl=pl, update_first=False), 'abc')
		self.assertEqual(thread_number(), 1)
		self.assertEqual(log, [
			('set_state', {}),
			('update', None),
			('render', 'abc', {'pl': pl, 'update_first': False}),
		])
		log[:] = ()

		segment = TestSegment()
		block_event.set()
		updates[0] = ValueError('abc')
		self.assertEqual(segment(pl=pl), None)
		self.assertEqual(thread_number(), 1)
		self.assertEqual(len(pl.exceptions), 1)
		self.assertEqual(log, [
			('set_state', {}),
			('update', None),
		])
		log[:] = ()
		pl.exceptions[:] = ()

		segment = TestSegment()
		block_event.set()
		updates[0] = (TypeError('def'),)
		self.assertRaises(TypeError, segment, pl=pl)
		self.assertEqual(thread_number(), 1)
		self.assertEqual(log, [
			('set_state', {}),
			('update', None),
			('render', updates[0][0], {'pl': pl, 'update_first': True}),
		])
		log[:] = ()

		# Threaded tests
		segment = TestSegment()
		block_event.clear()
		kwargs = {'pl': pl, 'update_first': False, 'other': 1}
		with lock:
			updates[0] = ('abc',)
		segment.startup(**kwargs)
		ret = segment(**kwargs)
		self.assertEqual(thread_number(), 2)
		block_event.set()
		event.wait()
		segment.shutdown_event.set()
		segment.thread.join()
		self.assertEqual(ret, None)
		self.assertEqual(log, [
			('set_state', {'update_first': False, 'other': 1}),
			('render', None, {'pl': pl, 'update_first': False, 'other': 1}),
			('update', None),
		])
		log[:] = ()

		segment = TestSegment()
		block_event.set()
		kwargs = {'pl': pl, 'update_first': True, 'other': 1}
		with lock:
			updates[0] = ('def',)
		segment.startup(**kwargs)
		ret = segment(**kwargs)
		self.assertEqual(thread_number(), 2)
		segment.shutdown_event.set()
		segment.thread.join()
		self.assertEqual(ret, 'def')
		self.assertEqual(log, [
			('set_state', {'update_first': True, 'other': 1}),
			('update', None),
			('render', 'def', {'pl': pl, 'update_first': True, 'other': 1}),
		])
		log[:] = ()

		segment = TestSegment()
		block_event.set()
		kwargs = {'pl': pl, 'update_first': True, 'interval': 0.2}
		with lock:
			updates[0] = ('abc',)
		segment.startup(**kwargs)
		start = monotonic()
		ret1 = segment(**kwargs)
		with lock:
			updates[0] = ('def',)
		self.assertEqual(thread_number(), 2)
		sleep(0.5)
		ret2 = segment(**kwargs)
		segment.shutdown_event.set()
		segment.thread.join()
		end = monotonic()
		duration = end - start
		self.assertEqual(ret1, 'abc')
		self.assertEqual(ret2, 'def')
		self.assertEqual(log[:5], [
			('set_state', {'update_first': True, 'interval': 0.2}),
			('update', None),
			('render', 'abc', {'pl': pl, 'update_first': True, 'interval': 0.2}),
			('update', 'abc'),
			('update', 'def'),
		])
		num_runs = len([e for e in log if e[0] == 'update'])
		self.assertAlmostEqual(duration / 0.2, num_runs, delta=1)
		log[:] = ()

		segment = TestSegment()
		block_event.set()
		kwargs = {'pl': pl, 'update_first': True, 'interval': 0.2}
		with lock:
			updates[0] = ('ghi',)
		segment.startup(**kwargs)
		start = monotonic()
		ret1 = segment(**kwargs)
		with lock:
			updates[0] = TypeError('jkl')
		self.assertEqual(thread_number(), 2)
		sleep(0.5)
		ret2 = segment(**kwargs)
		segment.shutdown_event.set()
		segment.thread.join()
		end = monotonic()
		duration = end - start
		self.assertEqual(ret1, 'ghi')
		self.assertEqual(ret2, None)
		self.assertEqual(log[:5], [
			('set_state', {'update_first': True, 'interval': 0.2}),
			('update', None),
			('render', 'ghi', {'pl': pl, 'update_first': True, 'interval': 0.2}),
			('update', 'ghi'),
			('update', 'ghi'),
		])
		num_runs = len([e for e in log if e[0] == 'update'])
		self.assertAlmostEqual(duration / 0.2, num_runs, delta=1)
		self.assertEqual(num_runs - 1, len(pl.exceptions))
		log[:] = ()

	def test_kw_threaded_segment(self):
		log = []
		pl = Pl()
		event = threading.Event()

		class TestSegment(KwThreadedSegment):
			interval = 10

			@staticmethod
			def key(_key=(None,), **kwargs):
				log.append(('key', _key, kwargs))
				return _key

			def compute_state(self, key):
				event.set()
				sleep(0.1)
				log.append(('compute_state', key))
				ret = key
				if isinstance(ret, Exception):
					raise ret
				else:
					return ret[0]

			def render_one(self, state, **kwargs):
				log.append(('render_one', state, kwargs))
				if isinstance(state, Exception):
					raise state
				else:
					return state

		# Non-threaded tests
		segment = TestSegment()
		event.clear()
		self.assertEqual(segment(pl=pl), None)
		self.assertEqual(thread_number(), 1)
		self.assertEqual(log, [
			('key', (None,), {'pl': pl}),
			('compute_state', (None,)),
			('render_one', None, {'pl': pl}),
		])
		log[:] = ()

		segment = TestSegment()
		kwargs = {'pl': pl, '_key': ('abc',), 'update_first': False}
		event.clear()
		self.assertEqual(segment(**kwargs), 'abc')
		kwargs.update(_key=('def',))
		self.assertEqual(segment(**kwargs), 'def')
		self.assertEqual(thread_number(), 1)
		self.assertEqual(log, [
			('key', ('abc',), {'pl': pl}),
			('compute_state', ('abc',)),
			('render_one', 'abc', {'pl': pl, '_key': ('abc',)}),
			('key', ('def',), {'pl': pl}),
			('compute_state', ('def',)),
			('render_one', 'def', {'pl': pl, '_key': ('def',)}),
		])
		log[:] = ()

		segment = TestSegment()
		kwargs = {'pl': pl, '_key': ValueError('xyz'), 'update_first': False}
		event.clear()
		self.assertEqual(segment(**kwargs), None)
		self.assertEqual(thread_number(), 1)
		self.assertEqual(log, [
			('key', kwargs['_key'], {'pl': pl}),
			('compute_state', kwargs['_key']),
		])
		log[:] = ()

		segment = TestSegment()
		kwargs = {'pl': pl, '_key': (ValueError('abc'),), 'update_first': False}
		event.clear()
		self.assertRaises(ValueError, segment, **kwargs)
		self.assertEqual(thread_number(), 1)
		self.assertEqual(log, [
			('key', kwargs['_key'], {'pl': pl}),
			('compute_state', kwargs['_key']),
			('render_one', kwargs['_key'][0], {'pl': pl, '_key': kwargs['_key']}),
		])
		log[:] = ()

		# Threaded tests
		segment = TestSegment()
		kwargs = {'pl': pl, 'update_first': False, '_key': ('_abc',)}
		event.clear()
		segment.startup(**kwargs)
		ret = segment(**kwargs)
		self.assertEqual(thread_number(), 2)
		segment.shutdown_event.set()
		segment.thread.join()
		self.assertEqual(ret, None)
		self.assertEqual(log[:2], [
			('key', kwargs['_key'], {'pl': pl}),
			('render_one', None, {'pl': pl, '_key': kwargs['_key']}),
		])
		self.assertLessEqual(len(log), 3)
		if len(log) > 2:
			self.assertEqual(log[2], ('compute_state', kwargs['_key']))
		log[:] = ()

		segment = TestSegment()
		kwargs = {'pl': pl, 'update_first': True, '_key': ('_abc',)}
		event.clear()
		segment.startup(**kwargs)
		ret1 = segment(**kwargs)
		kwargs.update(_key=('_def',))
		ret2 = segment(**kwargs)
		self.assertEqual(thread_number(), 2)
		segment.shutdown_event.set()
		segment.thread.join()
		self.assertEqual(ret1, '_abc')
		self.assertEqual(ret2, '_def')
		self.assertEqual(log, [
			('key', ('_abc',), {'pl': pl}),
			('compute_state', ('_abc',)),
			('render_one', '_abc', {'pl': pl, '_key': ('_abc',)}),
			('key', ('_def',), {'pl': pl}),
			('compute_state', ('_def',)),
			('render_one', '_def', {'pl': pl, '_key': ('_def',)}),
		])
		log[:] = ()


class TestLib(TestCase):
	def test_mergedicts(self):
		d = {}
		mergedicts(d, {'abc': {'def': 'ghi'}})
		self.assertEqual(d, {'abc': {'def': 'ghi'}})
		mergedicts(d, {'abc': {'def': {'ghi': 'jkl'}}})
		self.assertEqual(d, {'abc': {'def': {'ghi': 'jkl'}}})
		mergedicts(d, {})
		self.assertEqual(d, {'abc': {'def': {'ghi': 'jkl'}}})
		mergedicts(d, {'abc': {'mno': 'pqr'}})
		self.assertEqual(d, {'abc': {'def': {'ghi': 'jkl'}, 'mno': 'pqr'}})
		mergedicts(d, {'abc': {'def': REMOVE_THIS_KEY}})
		self.assertEqual(d, {'abc': {'mno': 'pqr'}})

	def test_add_divider_highlight_group(self):
		def decorated_function_name(**kwargs):
			return str(kwargs)
		func = add_divider_highlight_group('hl_group')(decorated_function_name)
		self.assertEqual(func.__name__, 'decorated_function_name')
		self.assertEqual(func(kw={}), [{'contents': repr({str('kw'): {}}), 'divider_highlight_group': 'hl_group'}])

	def test_humanize_bytes(self):
		self.assertEqual(humanize_bytes(0), '0 B')
		self.assertEqual(humanize_bytes(1), '1 B')
		self.assertEqual(humanize_bytes(1, suffix='bit'), '1 bit')
		self.assertEqual(humanize_bytes(1000, si_prefix=True), '1 kB')
		self.assertEqual(humanize_bytes(1024, si_prefix=True), '1 kB')
		self.assertEqual(humanize_bytes(1000000000, si_prefix=True), '1.00 GB')
		self.assertEqual(humanize_bytes(1000000000, si_prefix=False), '953.7 MiB')


width_data = {
	'N': 1,          # Neutral
	'Na': 1,         # Narrow
	'A': 1,          # Ambiguous
	'H': 1,          # Half-width
	'W': 2,          # Wide
	'F': 2,          # Fullwidth
}


class TestUnicode(TestCase):
	def assertStringsIdentical(self, s1, s2):
		self.assertTrue(type(s1) is type(s2), msg='string types differ')
		self.assertEqual(s1, s2)

	def test_unicode(self):
		self.assertTrue(type('abc') is plu.unicode)

	def test_unichr(self):
		self.assertStringsIdentical('\U0010FFFF', plu.unichr(0x10FFFF))
		self.assertStringsIdentical('\uFFFF', plu.unichr(0xFFFF))
		self.assertStringsIdentical('\x20', plu.unichr(0x20))

	def test_u(self):
		self.assertStringsIdentical('Test', plu.u('Test'))
		self.assertStringsIdentical('Test', plu.u(b'Test'))
		self.assertStringsIdentical('«»', plu.u(b'\xC2\xAB\xC2\xBB'))
		self.assertRaises(UnicodeDecodeError, plu.u, b'\xFF')

	def test_tointiter(self):
		self.assertEqual([1, 2, 3], list(plu.tointiter(b'\x01\x02\x03')))

	def test_decode_error(self):
		self.assertStringsIdentical('<FF>', b'\xFF'.decode('utf-8', 'powerline_decode_error'))
		self.assertStringsIdentical('abc', b'abc'.decode('utf-8', 'powerline_decode_error'))

	def test_register_strwidth_error(self):
		ename = plu.register_strwidth_error(lambda s: 3)
		self.assertStringsIdentical(b'???', 'Ａ'.encode('latin1', ename))
		self.assertStringsIdentical(b'abc', 'abc'.encode('latin1', ename))

	def test_out_u(self):
		self.assertStringsIdentical('abc', plu.out_u('abc'))
		self.assertStringsIdentical('abc', plu.out_u(b'abc'))
		self.assertRaises(TypeError, plu.out_u, None)

	def test_safe_unicode(self):
		self.assertStringsIdentical('abc', plu.safe_unicode('abc'))
		self.assertStringsIdentical('abc', plu.safe_unicode(b'abc'))
		self.assertStringsIdentical('«»', plu.safe_unicode(b'\xc2\xab\xc2\xbb'))
		with replace_attr(plu, 'get_preferred_output_encoding', lambda: 'latin1'):
			self.assertStringsIdentical('ÿ', plu.safe_unicode(b'\xFF'))
		self.assertStringsIdentical('None', plu.safe_unicode(None))

		class FailingStr(object):
			def __str__(self):
				raise NotImplementedError('Fail!')

		self.assertStringsIdentical('Fail!', plu.safe_unicode(FailingStr()))

	def test_FailedUnicode(self):
		self.assertTrue(isinstance(plu.FailedUnicode('abc'), plu.unicode))
		self.assertEqual('abc', plu.FailedUnicode('abc'))

	def test_string(self):
		self.assertStringsIdentical(str('abc'), plu.string('abc'))
		self.assertStringsIdentical(str('abc'), plu.string(b'abc'))

	def test_surrogate_pair_to_character(self):
		self.assertEqual(0x1F48E, plu.surrogate_pair_to_character(0xD83D, 0xDC8E))

	def test_strwidth_ucs_4(self):
		self.assertEqual(4, plu.strwidth_ucs_4(width_data, 'abcd'))
		self.assertEqual(4, plu.strwidth_ucs_4(width_data, 'ＡＢ'))
		if sys.maxunicode < 0x10FFFF:
			raise SkipTest('Can only test strwidth_ucs_4 in UCS-4 Pythons')

		self.assertEqual(1, plu.strwidth_ucs_4(width_data, '\U0001F063'))

	def test_strwidth_ucs_2(self):
		self.assertEqual(4, plu.strwidth_ucs_2(width_data, 'abcd'))
		self.assertEqual(4, plu.strwidth_ucs_2(width_data, 'ＡＢ'))
		if not sys.maxunicode < 0x10FFFF:
			raise SkipTest('Can only test strwidth_ucs_2 in UCS-2 Pythons')
		self.assertEqual(1, plu.strwidth_ucs_2(width_data, '\ud83c\udc30'))


class TestVCS(TestCase):
	def do_branch_rename_test(self, repo, q):
		st = monotonic()
		while monotonic() - st < 1:
			# Give inotify time to deliver events
			ans = repo.branch()
			if hasattr(q, '__call__'):
				if q(ans):
					break
			else:
				if ans == q:
					break
			sleep(0.01)
		if hasattr(q, '__call__'):
			self.assertTrue(q(ans))
		else:
			self.assertEqual(ans, q)

	def test_git(self):
		create_watcher = get_fallback_create_watcher()
		repo = guess(path=GIT_REPO, create_watcher=create_watcher)
		self.assertNotEqual(repo, None)
		self.assertEqual(repo.branch(), 'master')
		self.assertEqual(repo.status(), None)
		self.assertEqual(repo.status('file'), None)
		with open(os.path.join(GIT_REPO, 'file'), 'w') as f:
			f.write('abc')
			f.flush()
			self.assertEqual(repo.status(), '  U')
			self.assertEqual(repo.status('file'), '??')
			call(['git', 'add', '.'], cwd=GIT_REPO)
			self.assertEqual(repo.status(), ' I ')
			self.assertEqual(repo.status('file'), 'A ')
			f.write('def')
			f.flush()
			self.assertEqual(repo.status(), 'DI ')
			self.assertEqual(repo.status('file'), 'AM')
		os.remove(os.path.join(GIT_REPO, 'file'))
		# Test changing branch
		self.assertEqual(repo.branch(), 'master')
		try:
			call(['git', 'branch', 'branch1'], cwd=GIT_REPO)
			call(['git', 'checkout', '-q', 'branch1'], cwd=GIT_REPO)
			self.do_branch_rename_test(repo, 'branch1')
			call(['git', 'branch', 'branch2'], cwd=GIT_REPO)
			call(['git', 'checkout', '-q', 'branch2'], cwd=GIT_REPO)
			self.do_branch_rename_test(repo, 'branch2')
			call(['git', 'checkout', '-q', '--detach', 'branch1'], cwd=GIT_REPO)
			self.do_branch_rename_test(repo, lambda b: re.match(r'^[a-f0-9]+$', b))
		finally:
			call(['git', 'checkout', '-q', 'master'], cwd=GIT_REPO)
		# Test stashing
		self.assertEqual(repo.stash(), 0)

		def stash_save():
			with open(os.path.join(GIT_REPO, 'file'), 'w') as f:
				f.write('abc')
			return call(['git', 'stash', '-u'], cwd=GIT_REPO, stdout=PIPE)

		def stash_drop():
			return call(['git', 'stash', 'drop'], cwd=GIT_REPO, stdout=PIPE)

		def stash_list():
			return call(['git', 'stash', 'list'], cwd=GIT_REPO, stdout=PIPE)

		try:
			stash_save()
			self.assertEqual(repo.stash(), 1)
			stash_save()
			self.assertEqual(repo.stash(), 2)
			stash_drop()
			self.assertEqual(repo.stash(), 1)
			stash_drop()
			self.assertEqual(repo.stash(), 0)
		finally:
			while stash_list():
			    stash_drop()

	def test_git_sym(self):
		create_watcher = get_fallback_create_watcher()
		dotgit = os.path.join(GIT_REPO, '.git')
		spacegit = os.path.join(GIT_REPO, ' .git ')
		os.rename(dotgit, spacegit)
		try:
			with open(dotgit, 'w') as F:
				F.write('gitdir:  .git \n')
			gitdir = git_directory(GIT_REPO)
			self.assertTrue(os.path.isdir(gitdir))
			self.assertEqual(gitdir, os.path.abspath(spacegit))
			repo = guess(path=GIT_REPO, create_watcher=create_watcher)
			self.assertEqual(repo.branch(), 'master')
		finally:
			os.remove(dotgit)
			os.rename(spacegit, dotgit)

	def test_mercurial(self):
		if not use_mercurial:
			raise SkipTest('Mercurial is not available')
		create_watcher = get_fallback_create_watcher()
		repo = guess(path=HG_REPO, create_watcher=create_watcher)
		self.assertNotEqual(repo, None)
		self.assertEqual(repo.branch(), 'default')
		self.assertEqual(repo.status(), None)
		with open(os.path.join(HG_REPO, 'file'), 'w') as f:
			f.write('abc')
			f.flush()
			self.assertEqual(repo.status(), ' U')
			self.assertEqual(repo.status('file'), 'U')
			call(['hg', 'add', '.'], cwd=HG_REPO, stdout=PIPE)
			self.assertEqual(repo.status(), 'D ')
			self.assertEqual(repo.status('file'), 'A')
		os.remove(os.path.join(HG_REPO, 'file'))

	def test_bzr(self):
		if not use_bzr:
			raise SkipTest('Bazaar is not available')
		create_watcher = get_fallback_create_watcher()
		repo = guess(path=BZR_REPO, create_watcher=create_watcher)
		self.assertNotEqual(repo, None, 'No bzr repo found. Do you have bzr installed?')
		self.assertEqual(repo.branch(), 'test_powerline')
		self.assertEqual(repo.status(), None)
		with open(os.path.join(BZR_REPO, 'file'), 'w') as f:
			f.write('abc')
		self.assertEqual(repo.status(), ' U')
		self.assertEqual(repo.status('file'), '? ')
		call(['bzr', 'add', '-q', '.'], cwd=BZR_REPO, stdout=PIPE)
		self.assertEqual(repo.status(), 'D ')
		self.assertEqual(repo.status('file'), '+N')
		call(['bzr', 'commit', '-q', '-m', 'initial commit'], cwd=BZR_REPO)
		self.assertEqual(repo.status(), None)
		with open(os.path.join(BZR_REPO, 'file'), 'w') as f:
			f.write('def')
		self.assertEqual(repo.status(), 'D ')
		self.assertEqual(repo.status('file'), ' M')
		self.assertEqual(repo.status('notexist'), None)
		with open(os.path.join(BZR_REPO, 'ignored'), 'w') as f:
			f.write('abc')
		self.assertEqual(repo.status('ignored'), '? ')
		# Test changing the .bzrignore file should update status
		with open(os.path.join(BZR_REPO, '.bzrignore'), 'w') as f:
			f.write('ignored')
		self.assertEqual(repo.status('ignored'), None)
		# Test changing the dirstate file should invalidate the cache for
		# all files in the repo
		with open(os.path.join(BZR_REPO, 'file2'), 'w') as f:
			f.write('abc')
		call(['bzr', 'add', 'file2'], cwd=BZR_REPO, stdout=PIPE)
		call(['bzr', 'commit', '-q', '-m', 'file2 added'], cwd=BZR_REPO)
		with open(os.path.join(BZR_REPO, 'file'), 'a') as f:
			f.write('hello')
		with open(os.path.join(BZR_REPO, 'file2'), 'a') as f:
			f.write('hello')
		self.assertEqual(repo.status('file'), ' M')
		self.assertEqual(repo.status('file2'), ' M')
		call(['bzr', 'commit', '-q', '-m', 'multi'], cwd=BZR_REPO)
		self.assertEqual(repo.status('file'), None)
		self.assertEqual(repo.status('file2'), None)

		# Test changing branch
		call(['bzr', 'nick', 'branch1'], cwd=BZR_REPO, stdout=PIPE, stderr=PIPE)
		self.do_branch_rename_test(repo, 'branch1')

		# Test branch name/status changes when swapping repos
		for x in ('b1', 'b2'):
			d = os.path.join(BZR_REPO, x)
			os.mkdir(d)
			call(['bzr', 'init', '-q'], cwd=d)
			call(['bzr', 'nick', '-q', x], cwd=d)
			repo = guess(path=d, create_watcher=create_watcher)
			self.assertEqual(repo.branch(), x)
			self.assertFalse(repo.status())
			if x == 'b1':
				open(os.path.join(d, 'dirty'), 'w').close()
				self.assertTrue(repo.status())
		os.rename(os.path.join(BZR_REPO, 'b1'), os.path.join(BZR_REPO, 'b'))
		os.rename(os.path.join(BZR_REPO, 'b2'), os.path.join(BZR_REPO, 'b1'))
		os.rename(os.path.join(BZR_REPO, 'b'), os.path.join(BZR_REPO, 'b2'))
		for x, y in (('b1', 'b2'), ('b2', 'b1')):
			d = os.path.join(BZR_REPO, x)
			repo = guess(path=d, create_watcher=create_watcher)
			self.do_branch_rename_test(repo, y)
			if x == 'b1':
				self.assertFalse(repo.status())
			else:
				self.assertTrue(repo.status())

	@classmethod
	def setUpClass(cls):
		cls.powerline_old_cwd = os.getcwd()
		os.chdir(os.path.dirname(os.path.dirname(__file__)))
		call(['git', 'init', '--quiet', GIT_REPO])
		assert os.path.isdir(GIT_REPO)
		call(['git', 'config', '--local', 'user.name', 'Foo'], cwd=GIT_REPO)
		call(['git', 'config', '--local', 'user.email', 'bar@example.org'], cwd=GIT_REPO)
		call(['git', 'commit', '--allow-empty', '--message', 'Initial commit', '--quiet'], cwd=GIT_REPO)
		if use_mercurial:
			cls.powerline_old_HGRCPATH = os.environ.get('HGRCPATH')
			os.environ['HGRCPATH'] = ''
			call(['hg', 'init', HG_REPO])
			with open(os.path.join(HG_REPO, '.hg', 'hgrc'), 'w') as hgrc:
				hgrc.write('[ui]\n')
				hgrc.write('username = Foo <bar@example.org>\n')
		if use_bzr:
			call(['bzr', 'init', '--quiet', BZR_REPO])
			call(['bzr', 'config', 'email=Foo <bar@example.org>'], cwd=BZR_REPO)
			call(['bzr', 'config', 'nickname=test_powerline'], cwd=BZR_REPO)
			call(['bzr', 'config', 'create_signatures=0'], cwd=BZR_REPO)

	@classmethod
	def tearDownClass(cls):
		for repo_dir in [GIT_REPO] + ([HG_REPO] if use_mercurial else []) + ([BZR_REPO] if use_bzr else []):
			shutil.rmtree(repo_dir)
		if use_mercurial:
			if cls.powerline_old_HGRCPATH is None:
				os.environ.pop('HGRCPATH')
			else:
				os.environ['HGRCPATH'] = cls.powerline_old_HGRCPATH
		os.chdir(cls.powerline_old_cwd)


if __name__ == '__main__':
	from tests.modules import main
	main()



# vim:fileencoding=utf-8:noet

'''Tests for various logging features'''

from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import re
import codecs
import os

from io import StringIO
from shutil import rmtree

from powerline import finish_common_config, create_logger

from tests.modules import TestCase
from tests.modules.lib import replace_attr


TIMESTAMP_RE = r'\d{4}-\d\d-\d\d \d\d:\d\d:\d\d,\d{3}'


class TestRE(TestCase):
	def assertMatches(self, text, regexp):
		self.assertTrue(
			re.match(regexp, text),
			'{0!r} did not match {1!r}'.format(text, regexp),
		)


def close_handlers(logger):
	for handler in logger.handlers:
		handler.close()


class TestHandlers(TestRE):
	def test_stderr_handler_is_default(self):
		out = StringIO()
		err = StringIO()

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {})
			logger, pl, get_module_attr = create_logger(common_config)
			pl.error('Foo')
			close_handlers(logger)
			self.assertMatches(err.getvalue(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
			self.assertEqual(out.getvalue(), '')

	def test_stream_override(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {})
			logger, pl, get_module_attr = create_logger(common_config, stream=stream)
			pl.error('Foo')
			close_handlers(logger)
			self.assertMatches(stream.getvalue(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_explicit_none(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [None]})
			logger, pl, get_module_attr = create_logger(common_config, stream=stream)
			pl.error('Foo')
			close_handlers(logger)
			self.assertMatches(stream.getvalue(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_explicit_stream_handler(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [['logging.StreamHandler', [[]]]]})
			logger, pl, get_module_attr = create_logger(common_config, stream=stream)
			pl.error('Foo')
			close_handlers(logger)
			self.assertEqual(stream.getvalue(), '')
			self.assertMatches(err.getvalue(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
			self.assertEqual(out.getvalue(), '')

	def test_explicit_stream_handler_implicit_stream(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [['logging.StreamHandler', []]]})
			logger, pl, get_module_attr = create_logger(common_config, stream=stream)
			pl.error('Foo')
			close_handlers(logger)
			self.assertMatches(stream.getvalue(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_file_handler(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()
		file_name = 'test_logging-test_file_handler'

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': file_name})
			try:
				logger, pl, get_module_attr = create_logger(common_config, stream=stream)
				pl.error('Foo')
				close_handlers(logger)
				with codecs.open(file_name, encoding='utf-8') as fp:
					self.assertMatches(fp.read(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
			finally:
				os.unlink(file_name)
			self.assertEqual(stream.getvalue(), '')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_file_handler_create_dir(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()
		file_name = 'test_logging-test_file_handler_create_dir/file'

		self.assertFalse(os.path.isdir(os.path.dirname(file_name)))

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': file_name})
			try:
				logger, pl, get_module_attr = create_logger(common_config, stream=stream)
				pl.error('Foo')
				close_handlers(logger)
				self.assertTrue(os.path.isdir(os.path.dirname(file_name)))
				with codecs.open(file_name, encoding='utf-8') as fp:
					self.assertMatches(fp.read(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
			finally:
				rmtree(os.path.dirname(file_name))
			self.assertEqual(stream.getvalue(), '')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_multiple_files(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()
		file_name_1 = 'test_logging-test_multiple_files-1'
		file_name_2 = file_name_1[:-1] + '2'

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [file_name_1, file_name_2]})
			try:
				try:
					logger, pl, get_module_attr = create_logger(common_config, stream=stream)
					pl.error('Foo')
					close_handlers(logger)
					for file_name in (file_name_1, file_name_2):
						with codecs.open(file_name, encoding='utf-8') as fp:
							self.assertMatches(fp.read(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
				finally:
					os.unlink(file_name_1)
			finally:
				os.unlink(file_name_2)
			self.assertEqual(stream.getvalue(), '')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_multiple_files_and_stream(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()
		file_name_1 = 'test_logging-test_multiple_files_and_stream-1'
		file_name_2 = file_name_1[:-1] + '2'

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [file_name_1, file_name_2, None]})
			try:
				try:
					logger, pl, get_module_attr = create_logger(common_config, stream=stream)
					pl.error('Foo')
					close_handlers(logger)
					for file_name in (file_name_1, file_name_2):
						with codecs.open(file_name, encoding='utf-8') as fp:
							self.assertMatches(fp.read(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
				finally:
					os.unlink(file_name_1)
			finally:
				os.unlink(file_name_2)
			self.assertMatches(stream.getvalue(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_handler_args(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()
		file_name = 'test_logging-test_handler_args'

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [
				['RotatingFileHandler', [[file_name]]]
			]})
			try:
				logger, pl, get_module_attr = create_logger(common_config, stream=stream)
				pl.error('Foo')
				close_handlers(logger)
				with codecs.open(file_name, encoding='utf-8') as fp:
					self.assertMatches(fp.read(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
			finally:
				os.unlink(file_name)
			self.assertEqual(stream.getvalue(), '')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_handler_args_kwargs(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()
		file_name = 'test_logging-test_handler_args_kwargs'

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [
				['RotatingFileHandler', [[file_name], {'maxBytes': 1, 'backupCount': 1}]]
			]})
			try:
				try:
					logger, pl, get_module_attr = create_logger(common_config, stream=stream)
					pl.error('Foo')
					pl.error('Bar')
					close_handlers(logger)
					with codecs.open(file_name, encoding='utf-8') as fp:
						self.assertMatches(fp.read(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Bar\n$')
					with codecs.open(file_name + '.1', encoding='utf-8') as fp:
						self.assertMatches(fp.read(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Foo\n$')
				finally:
					os.unlink(file_name + '.1')
			finally:
				os.unlink(file_name)
			self.assertEqual(stream.getvalue(), '')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_logger_level(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()
		stream1 = StringIO()
		stream2 = StringIO()

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [
				['logging.StreamHandler', [[stream1]], 'WARNING'],
				['logging.StreamHandler', [[stream2]], 'ERROR'],
			]})
			logger, pl, get_module_attr = create_logger(common_config, stream=stream)
			pl.warn('Foo')
			pl.error('Bar')
			close_handlers(logger)
			self.assertMatches(stream1.getvalue(), (
				'^' + TIMESTAMP_RE + ':WARNING:__unknown__:Foo\n'
				+ TIMESTAMP_RE + ':ERROR:__unknown__:Bar\n$'
			))
			self.assertMatches(stream2.getvalue(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Bar\n$')
			self.assertEqual(stream.getvalue(), '')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_logger_level_not_overriding_default(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()
		stream1 = StringIO()

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [
				['logging.StreamHandler', [[stream1]], 'DEBUG'],
			]})
			logger, pl, get_module_attr = create_logger(common_config, stream=stream)
			pl.debug('Foo')
			pl.error('Bar')
			close_handlers(logger)
			self.assertMatches(stream1.getvalue(), '^' + TIMESTAMP_RE + ':ERROR:__unknown__:Bar\n$')
			self.assertEqual(stream.getvalue(), '')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_top_log_level(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()
		stream1 = StringIO()

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [
				['logging.StreamHandler', [[stream1]], 'DEBUG'],
			], 'log_level': 'DEBUG'})
			logger, pl, get_module_attr = create_logger(common_config, stream=stream)
			pl.debug('Foo')
			pl.error('Bar')
			close_handlers(logger)
			self.assertMatches(stream1.getvalue(), (
				'^' + TIMESTAMP_RE + ':DEBUG:__unknown__:Foo\n'
				+ TIMESTAMP_RE + ':ERROR:__unknown__:Bar\n$'
			))
			self.assertEqual(stream.getvalue(), '')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_logger_format(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()
		stream1 = StringIO()

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [
				['logging.StreamHandler', [[stream1]], 'WARNING', 'FOO'],
			]})
			logger, pl, get_module_attr = create_logger(common_config, stream=stream)
			pl.warn('Foo')
			pl.error('Bar')
			close_handlers(logger)
			self.assertEqual(stream1.getvalue(), 'FOO\nFOO\n')
			self.assertEqual(stream.getvalue(), '')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')

	def test_top_log_format(self):
		out = StringIO()
		err = StringIO()
		stream = StringIO()
		stream1 = StringIO()
		stream2 = StringIO()

		with replace_attr(sys, 'stdout', out, 'stderr', err):
			common_config = finish_common_config('utf-8', {'log_file': [
				['logging.StreamHandler', [[stream1]], 'WARNING', 'FOO'],
				['logging.StreamHandler', [[stream2]], 'WARNING'],
			], 'log_format': 'BAR'})
			logger, pl, get_module_attr = create_logger(common_config, stream=stream)
			pl.warn('Foo')
			pl.error('Bar')
			close_handlers(logger)
			self.assertEqual(stream2.getvalue(), 'BAR\nBAR\n')
			self.assertEqual(stream1.getvalue(), 'FOO\nFOO\n')
			self.assertEqual(stream.getvalue(), '')
			self.assertEqual(err.getvalue(), '')
			self.assertEqual(out.getvalue(), '')


class TestPowerlineLogger(TestRE):
	def test_args_formatting(self):
		stream = StringIO()

		common_config = finish_common_config('utf-8', {})
		logger, pl, get_module_attr = create_logger(common_config, stream=stream)
		pl.warn('foo {0}', 'Test')
		pl.warn('bar {0!r}', 'Test')
		close_handlers(logger)
		self.assertMatches(stream.getvalue(), (
			'^' + TIMESTAMP_RE + ':WARNING:__unknown__:foo Test\n'
			+ TIMESTAMP_RE + ':WARNING:__unknown__:bar u?\'Test\'\n$'
		))

	def test_prefix_formatting(self):
		stream = StringIO()

		common_config = finish_common_config('utf-8', {})
		logger, pl, get_module_attr = create_logger(common_config, stream=stream)
		pl.prefix = '1'
		pl.warn('foo')
		pl.prefix = '2'
		pl.warn('bar')
		close_handlers(logger)
		self.assertMatches(stream.getvalue(), (
			'^' + TIMESTAMP_RE + ':WARNING:__unknown__:1:foo\n'
			+ TIMESTAMP_RE + ':WARNING:__unknown__:2:bar\n$'
		))

	def test_kwargs_formatting(self):
		stream = StringIO()

		common_config = finish_common_config('utf-8', {})
		logger, pl, get_module_attr = create_logger(common_config, stream=stream)
		pl.warn('foo {arg}', arg='Test')
		pl.warn('bar {arg!r}', arg='Test')
		close_handlers(logger)
		self.assertMatches(stream.getvalue(), (
			'^' + TIMESTAMP_RE + ':WARNING:__unknown__:foo Test\n'
			+ TIMESTAMP_RE + ':WARNING:__unknown__:bar u?\'Test\'\n$'
		))

	def test_args_kwargs_formatting(self):
		stream = StringIO()

		common_config = finish_common_config('utf-8', {})
		logger, pl, get_module_attr = create_logger(common_config, stream=stream)
		pl.warn('foo {0!r} {arg}', 'Test0', arg='Test')
		pl.warn('bar {0} {arg!r}', 'Test0', arg='Test')
		close_handlers(logger)
		self.assertMatches(stream.getvalue(), (
			'^' + TIMESTAMP_RE + ':WARNING:__unknown__:foo u?\'Test0\' Test\n'
			+ TIMESTAMP_RE + ':WARNING:__unknown__:bar Test0 u?\'Test\'\n$'
		))

	def test_exception_formatting(self):
		stream = StringIO()

		common_config = finish_common_config('utf-8', {})
		logger, pl, get_module_attr = create_logger(common_config, stream=stream)
		try:
			raise ValueError('foo')
		except ValueError:
			pl.exception('Message')
		close_handlers(logger)
		self.assertMatches(stream.getvalue(), (
			'^' + TIMESTAMP_RE + ':ERROR:__unknown__:Message\n'
			+ 'Traceback \\(most recent call last\\):\n'
			+ '(?:  File ".*?", line \\d+, in \\w+\n    [^\n]*\n)+'
			+ 'ValueError: foo\n$'
		))

	def test_levels(self):
		stream = StringIO()

		common_config = finish_common_config('utf-8', {'log_level': 'DEBUG'})
		logger, pl, get_module_attr = create_logger(common_config, stream=stream)
		pl.debug('1')
		pl.info('2')
		pl.warn('3')
		pl.error('4')
		pl.critical('5')
		close_handlers(logger)
		self.assertMatches(stream.getvalue(), (
			'^' + TIMESTAMP_RE + ':DEBUG:__unknown__:1\n'
			+ TIMESTAMP_RE + ':INFO:__unknown__:2\n'
			+ TIMESTAMP_RE + ':WARNING:__unknown__:3\n'
			+ TIMESTAMP_RE + ':ERROR:__unknown__:4\n'
			+ TIMESTAMP_RE + ':CRITICAL:__unknown__:5\n$'
		))


old_cwd = None


def setUpModule():
	global old_cwd
	global __file__
	old_cwd = os.getcwd()
	__file__ = os.path.abspath(__file__)
	os.chdir(os.path.dirname(os.path.dirname(__file__)))


def tearDownModule():
	global old_cwd
	os.chdir(old_cwd)


if __name__ == '__main__':
	from tests.modules import main
	main()



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os
import sys

from functools import partial

import tests.modules.vim as vim_module

from tests.modules.lib import Pl
from tests.modules import TestCase


class TestVim(TestCase):
	def test_single_tab(self):
		pl = Pl()
		single_tab = partial(self.vim.single_tab, pl=pl, segment_info=None, mode=None)
		with vim_module._with('tabpage'):
			self.assertEqual(single_tab(), False)
		self.assertEqual(single_tab(), True)

	@classmethod
	def setUpClass(cls):
		sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'vim_sys_path')))
		from powerline.selectors import vim
		cls.vim = vim

	@classmethod
	def tearDownClass(cls):
		sys.path.pop(0)


if __name__ == '__main__':
	from tests.modules import main
	main()



# vim:fileencoding=utf-8:noet
_log = []
vars = {}
vvars = {'version': 703}
_tabpage = 0
_mode = 'n'
_buf_purge_events = set()
options = {
	'paste': 0,
	'ambiwidth': 'single',
	'columns': 80,
	'encoding': 'utf-8',
}
_last_bufnr = 0
_highlights = {}
from collections import defaultdict as _defaultdict
_environ = _defaultdict(lambda: '')
del _defaultdict


_thread_id = None


def _set_thread_id():
	global _thread_id
	from threading import current_thread
	_thread_id = current_thread().ident


# Assuming import is done from the main thread
_set_thread_id()


def _print_log():
	for item in _log:
		print (item)
	_log[:] = ()


def _vim(func):
	from functools import wraps
	from threading import current_thread

	@wraps(func)
	def f(*args, **kwargs):
		global _thread_id
		if _thread_id != current_thread().ident:
			raise RuntimeError('Accessing vim from separate threads is not allowed')
		_log.append((func.__name__, args))
		return func(*args, **kwargs)

	return f


def _unicode(func):
	from functools import wraps
	import sys

	if sys.version_info < (3,):
		return func

	@wraps(func)
	def f(*args, **kwargs):
		from powerline.lib.unicode import u
		ret = func(*args, **kwargs)
		if isinstance(ret, bytes):
			ret = u(ret)
		return ret

	return f


class _Buffers(object):
	@_vim
	def __init__(self):
		self.d = {}

	@_vim
	def __len__(self):
		return len(self.d)

	@_vim
	def __getitem__(self, item):
		return self.d[item]

	@_vim
	def __setitem__(self, item, value):
		self.d[item] = value

	@_vim
	def __iter__(self):
		return iter(self.d.values())

	@_vim
	def __contains__(self, item):
		return item in self.d

	@_vim
	def _keys(self):
		return self.d.keys()

	@_vim
	def _pop(self, *args, **kwargs):
		return self.d.pop(*args, **kwargs)


buffers = _Buffers()


class _ObjList(object):
	@_vim
	def __init__(self, objtype):
		self.l = []
		self.objtype = objtype

	@_vim
	def __getitem__(self, item):
		return self.l[item - int(item > 0)]

	@_vim
	def __len__(self):
		return len(self.l)

	@_vim
	def __iter__(self):
		return iter(self.l)

	@_vim
	def _pop(self, idx):
		obj = self.l.pop(idx - 1)
		for moved_obj in self.l[idx - 1:]:
			moved_obj.number -= 1
		return obj

	@_vim
	def _append(self, *args, **kwargs):
		return self.l.append(*args, **kwargs)

	@_vim
	def _new(self, *args, **kwargs):
		number = len(self) + 1
		new_obj = self.objtype(number, *args, **kwargs)
		self._append(new_obj)
		return new_obj


def _construct_result(r):
	import sys
	if sys.version_info < (3,):
		return r
	else:
		if isinstance(r, str):
			return r.encode('utf-8')
		elif isinstance(r, list):
			return [_construct_result(i) for i in r]
		elif isinstance(r, dict):
			return dict((
				(_construct_result(k), _construct_result(v))
				for k, v in r.items()
			))
		return r


def _str_func(func):
	from functools import wraps

	@wraps(func)
	def f(*args, **kwargs):
		return _construct_result(func(*args, **kwargs))
	return f


def _log_print():
	import sys
	for entry in _log:
		sys.stdout.write(repr(entry) + '\n')


_current_group = None
_on_wipeout = []


@_vim
def command(cmd):
	global _current_group
	cmd = cmd.lstrip()
	if cmd.startswith('let g:'):
		import re
		varname, value = re.compile(r'^let g:(\w+)\s*=\s*(.*)').match(cmd).groups()
		vars[varname] = value
	elif cmd.startswith('hi '):
		sp = cmd.split()
		_highlights[sp[1]] = sp[2:]
	elif cmd.startswith('augroup'):
		augroup = cmd.partition(' ')[2]
		if augroup.upper() == 'END':
			_current_group = None
		else:
			_current_group = augroup
	elif cmd.startswith('autocmd'):
		rest = cmd.partition(' ')[2]
		auevent, rest = rest.partition(' ')[::2]
		pattern, aucmd = rest.partition(' ')[::2]
		if auevent != 'BufWipeout' or pattern != '*':
			raise NotImplementedError
		import sys
		if sys.version_info < (3,):
			if not aucmd.startswith(':python '):
				raise NotImplementedError
		else:
			if not aucmd.startswith(':python3 '):
				raise NotImplementedError
		_on_wipeout.append(aucmd.partition(' ')[2])
	elif cmd.startswith('set '):
		if cmd.startswith('set statusline='):
			options['statusline'] = cmd[len('set statusline='):]
		elif cmd.startswith('set tabline='):
			options['tabline'] = cmd[len('set tabline='):]
		else:
			raise NotImplementedError(cmd)
	else:
		raise NotImplementedError(cmd)


@_vim
@_unicode
def eval(expr):
	if expr.startswith('g:'):
		return vars[expr[2:]]
	elif expr.startswith('v:'):
		return vvars[expr[2:]]
	elif expr.startswith('&'):
		return options[expr[1:]]
	elif expr.startswith('$'):
		return _environ[expr[1:]]
	elif expr.startswith('PowerlineRegisterCachePurgerEvent'):
		_buf_purge_events.add(expr[expr.find('"') + 1:expr.rfind('"') - 1])
		return '0'
	elif expr.startswith('exists('):
		return '0'
	elif expr.startswith('getwinvar('):
		import re
		match = re.match(r'^getwinvar\((\d+), "(\w+)"\)$', expr)
		if not match:
			raise NotImplementedError(expr)
		winnr = int(match.group(1))
		varname = match.group(2)
		return _emul_getwinvar(winnr, varname)
	elif expr.startswith('has_key('):
		import re
		match = re.match(r'^has_key\(getwinvar\((\d+), ""\), "(\w+)"\)$', expr)
		if match:
			winnr = int(match.group(1))
			varname = match.group(2)
			return 0 + (varname in current.tabpage.windows[winnr].vars)
		else:
			match = re.match(r'^has_key\(gettabwinvar\((\d+), (\d+), ""\), "(\w+)"\)$', expr)
			if not match:
				raise NotImplementedError(expr)
			tabnr = int(match.group(1))
			winnr = int(match.group(2))
			varname = match.group(3)
			return 0 + (varname in tabpages[tabnr].windows[winnr].vars)
	elif expr == 'getbufvar("%", "NERDTreeRoot").path.str()':
		import os
		assert os.path.basename(current.buffer.name).startswith('NERD_tree_')
		return '/usr/include'
	elif expr.startswith('getbufvar('):
		import re
		match = re.match(r'^getbufvar\((\d+), ["\'](.+)["\']\)$', expr)
		if not match:
			raise NotImplementedError(expr)
		bufnr = int(match.group(1))
		varname = match.group(2)
		return _emul_getbufvar(bufnr, varname)
	elif expr == 'tabpagenr()':
		return current.tabpage.number
	elif expr == 'tabpagenr("$")':
		return len(tabpages)
	elif expr.startswith('tabpagewinnr('):
		tabnr = int(expr[len('tabpagewinnr('):-1])
		return tabpages[tabnr].window.number
	elif expr.startswith('tabpagebuflist('):
		import re
		match = re.match(r'tabpagebuflist\((\d+)\)\[(\d+)\]', expr)
		tabnr = int(match.group(1))
		winnr = int(match.group(2)) + 1
		return tabpages[tabnr].windows[winnr].buffer.number
	elif expr.startswith('gettabwinvar('):
		import re
		match = re.match(r'gettabwinvar\((\d+), (\d+), "(\w+)"\)', expr)
		tabnr = int(match.group(1))
		winnr = int(match.group(2))
		varname = match.group(3)
		return tabpages[tabnr].windows[winnr].vars[varname]
	elif expr.startswith('type(function('):
		import re
		match = re.match(r'^type\(function\("([^"]+)"\)\) == 2$', expr)
		if not match:
			raise NotImplementedError(expr)
		return 0
	raise NotImplementedError(expr)


@_vim
def bindeval(expr):
	if expr == 'g:':
		return vars
	elif expr == '{}':
		return {}
	elif expr == '[]':
		return []
	import re
	match = re.compile(r'^function\("([^"\\]+)"\)$').match(expr)
	if match:
		return globals()['_emul_' + match.group(1)]
	else:
		raise NotImplementedError


@_vim
@_str_func
def _emul_mode(*args):
	if args and args[0]:
		return _mode
	else:
		return _mode[0]


@_vim
@_str_func
def _emul_getbufvar(bufnr, varname):
	import re
	if varname[0] == '&':
		if bufnr == '%':
			bufnr = current.buffer.number
		if bufnr not in buffers:
			return ''
		try:
			return buffers[bufnr].options[varname[1:]]
		except KeyError:
			try:
				return options[varname[1:]]
			except KeyError:
				return ''
	elif re.match('^[a-zA-Z_]+$', varname):
		if bufnr == '%':
			bufnr = current.buffer.number
		if bufnr not in buffers:
			return ''
		return buffers[bufnr].vars[varname]
	raise NotImplementedError


@_vim
@_str_func
def _emul_getwinvar(winnr, varname):
	return current.tabpage.windows[winnr].vars.get(varname, '')


@_vim
def _emul_setwinvar(winnr, varname, value):
	current.tabpage.windows[winnr].vars[varname] = value


@_vim
def _emul_virtcol(expr):
	if expr == '.':
		return current.window.cursor[1] + 1
	if isinstance(expr, list) and len(expr) == 3:
		return expr[-2] + expr[-1]
	raise NotImplementedError


_v_pos = None


@_vim
def _emul_getpos(expr):
	if expr == '.':
		return [0, current.window.cursor[0] + 1, current.window.cursor[1] + 1, 0]
	if expr == 'v':
		return _v_pos or [0, current.window.cursor[0] + 1, current.window.cursor[1] + 1, 0]
	raise NotImplementedError


@_vim
@_str_func
def _emul_fnamemodify(path, modstring):
	import os
	_modifiers = {
		'~': lambda path: path.replace(os.environ['HOME'].encode('utf-8'), b'~') if path.startswith(os.environ['HOME'].encode('utf-8')) else path,
		'.': lambda path: (lambda tpath: path if tpath[:3] == b'..' + os.sep.encode() else tpath)(os.path.relpath(path)),
		't': lambda path: os.path.basename(path),
		'h': lambda path: os.path.dirname(path),
	}

	for mods in modstring.split(':')[1:]:
		path = _modifiers[mods](path)
	return path


@_vim
@_str_func
def _emul_expand(expr):
	global _abuf
	if expr == '<abuf>':
		return _abuf or current.buffer.number
	raise NotImplementedError


@_vim
def _emul_bufnr(expr):
	if expr == '$':
		return _last_bufnr
	raise NotImplementedError


@_vim
def _emul_exists(ident):
	if ident.startswith('g:'):
		return ident[2:] in vars
	elif ident.startswith(':'):
		return 0
	raise NotImplementedError


@_vim
def _emul_line2byte(line):
	buflines = current.buffer._buf_lines
	if line == len(buflines) + 1:
		return sum((len(s) for s in buflines)) + 1
	raise NotImplementedError


@_vim
def _emul_line(expr):
	cursorline = current.window.cursor[0] + 1
	numlines = len(current.buffer._buf_lines)
	if expr == 'w0':
		return max(cursorline - 5, 1)
	if expr == 'w$':
		return min(cursorline + 5, numlines)
	raise NotImplementedError


@_vim
@_str_func
def _emul_strtrans(s):
	# FIXME Do more replaces
	return s.replace(b'\xFF', b'<ff>')


@_vim
@_str_func
def _emul_bufname(bufnr):
	try:
		return buffers[bufnr]._name or b''
	except KeyError:
		return b''


_window_id = 0


class _Window(object):
	def __init__(self, number, buffer=None, cursor=(1, 0), width=80):
		global _window_id
		self.cursor = cursor
		self.width = width
		self.number = number
		if buffer:
			if type(buffer) is _Buffer:
				self.buffer = buffer
			else:
				self.buffer = _Buffer(**buffer)
		else:
			self.buffer = _Buffer()
		_window_id += 1
		self._window_id = _window_id
		self.options = {}
		self.vars = {
			'powerline_window_id': self._window_id,
		}

	def __repr__(self):
		return '<window ' + str(self.number - 1) + '>'


class _Tabpage(object):
	def __init__(self, number):
		self.windows = _ObjList(_Window)
		self.number = number

	def _new_window(self, **kwargs):
		self.window = self.windows._new(**kwargs)
		return self.window

	def _close_window(self, winnr, open_window=True):
		curwinnr = self.window.number
		win = self.windows._pop(winnr)
		if self.windows and winnr == curwinnr:
			self.window = self.windows[-1]
		elif open_window:
			current.tabpage._new_window()
		return win

	def _close(self):
		global _tabpage
		while self.windows:
			self._close_window(1, False)
		tabpages._pop(self.number)
		_tabpage = len(tabpages)


tabpages = _ObjList(_Tabpage)


_abuf = None


class _Buffer(object):
	def __init__(self, name=None):
		global _last_bufnr
		_last_bufnr += 1
		bufnr = _last_bufnr
		self.number = bufnr
		# FIXME Use unicode() for python-3
		self.name = name
		self.vars = {'changedtick': 1}
		self.options = {
			'modified': 0,
			'readonly': 0,
			'fileformat': 'unix',
			'filetype': '',
			'buftype': '',
			'fileencoding': 'utf-8',
			'textwidth': 80,
		}
		self._buf_lines = ['']
		self._undostate = [self._buf_lines[:]]
		self._undo_written = len(self._undostate)
		buffers[bufnr] = self

	@property
	def name(self):
		import sys
		if sys.version_info < (3,):
			return self._name
		else:
			return str(self._name, 'utf-8') if self._name else None

	@name.setter
	def name(self, name):
		if name is None:
			self._name = None
		else:
			import os
			if type(name) is not bytes:
				name = name.encode('utf-8')
			if b':/' in name:
				self._name = name
			else:
				self._name = os.path.abspath(name)

	def __getitem__(self, line):
		return self._buf_lines[line]

	def __setitem__(self, line, value):
		self.options['modified'] = 1
		self.vars['changedtick'] += 1
		self._buf_lines[line] = value
		from copy import copy
		self._undostate.append(copy(self._buf_lines))

	def __setslice__(self, *args):
		self.options['modified'] = 1
		self.vars['changedtick'] += 1
		self._buf_lines.__setslice__(*args)
		from copy import copy
		self._undostate.append(copy(self._buf_lines))

	def __getslice__(self, *args):
		return self._buf_lines.__getslice__(*args)

	def __len__(self):
		return len(self._buf_lines)

	def __repr__(self):
		return '<buffer ' + str(self.name) + '>'

	def __del__(self):
		global _abuf
		bufnr = self.number
		try:
			import __main__
		except ImportError:
			pass
		except RuntimeError:
			# Module may have already been garbage-collected
			pass
		else:
			if _on_wipeout:
				_abuf = bufnr
				try:
					for event in _on_wipeout:
						exec(event, __main__.__dict__)
				finally:
					_abuf = None


class _Current(object):
	@property
	def buffer(self):
		return self.window.buffer

	@property
	def window(self):
		return self.tabpage.window

	@property
	def tabpage(self):
		return tabpages[_tabpage - 1]


current = _Current()


_dict = None


@_vim
def _init():
	global _dict

	if _dict:
		return _dict

	_dict = {}
	for varname, value in globals().items():
		if varname[0] != '_':
			_dict[varname] = value
	_tabnew()
	return _dict


@_vim
def _get_segment_info():
	mode_translations = {
		chr(ord('V') - 0x40): '^V',
		chr(ord('S') - 0x40): '^S',
	}
	mode = _mode
	mode = mode_translations.get(mode, mode)
	window = current.window
	buffer = current.buffer
	tabpage = current.tabpage
	return {
		'window': window,
		'winnr': window.number,
		'buffer': buffer,
		'bufnr': buffer.number,
		'tabpage': tabpage,
		'tabnr': tabpage.number,
		'window_id': window._window_id,
		'mode': mode,
		'encoding': options['encoding'],
	}


@_vim
def _launch_event(event):
	pass


@_vim
def _start_mode(mode):
	global _mode
	if mode == 'i':
		_launch_event('InsertEnter')
	elif _mode == 'i':
		_launch_event('InsertLeave')
	_mode = mode


@_vim
def _undo():
	if len(current.buffer._undostate) == 1:
		return
	buffer = current.buffer
	buffer._undostate.pop(-1)
	buffer._buf_lines = buffer._undostate[-1]
	if buffer._undo_written == len(buffer._undostate):
		buffer.options['modified'] = 0


@_vim
def _edit(name=None):
	if current.buffer.name is None:
		buffer = current.buffer
		buffer.name = name
	else:
		buffer = _Buffer(name)
		current.window.buffer = buffer


@_vim
def _tabnew(name=None):
	global windows
	global _tabpage
	tabpage = tabpages._new()
	windows = tabpage.windows
	_tabpage = len(tabpages)
	_new(name)
	return tabpage


@_vim
def _new(name=None):
	current.tabpage._new_window(buffer={'name': name})


@_vim
def _split():
	current.tabpage._new_window(buffer=current.buffer)


@_vim
def _close(winnr, wipe=True):
	win = current.tabpage._close_window(winnr)
	if wipe:
		for w in current.tabpage.windows:
			if w.buffer.number == win.buffer.number:
				break
		else:
			_bw(win.buffer.number)


@_vim
def _bw(bufnr=None):
	bufnr = bufnr or current.buffer.number
	winnr = 1
	for win in current.tabpage.windows:
		if win.buffer.number == bufnr:
			_close(winnr, wipe=False)
		winnr += 1
	buffers._pop(bufnr)
	if not buffers:
		_Buffer()
	_b(max(buffers._keys()))


@_vim
def _b(bufnr):
	current.window.buffer = buffers[bufnr]


@_vim
def _set_cursor(line, col):
	current.window.cursor = (line, col)
	if _mode == 'n':
		_launch_event('CursorMoved')
	elif _mode == 'i':
		_launch_event('CursorMovedI')


@_vim
def _get_buffer():
	return current.buffer


@_vim
def _set_bufoption(option, value, bufnr=None):
	buffers[bufnr or current.buffer.number].options[option] = value
	if option == 'filetype':
		_launch_event('FileType')


class _WithNewBuffer(object):
	def __init__(self, func, *args, **kwargs):
		self.call = lambda: func(*args, **kwargs)

	def __enter__(self):
		self.call()
		self.bufnr = current.buffer.number
		return _get_segment_info()

	def __exit__(self, *args):
		_bw(self.bufnr)


@_vim
def _set_dict(d, new, setfunc=None):
	if not setfunc:
		def setfunc(k, v):
			d[k] = v

	old = {}
	na = []
	for k, v in new.items():
		try:
			old[k] = d[k]
		except KeyError:
			na.append(k)
		setfunc(k, v)
	return old, na


class _WithBufOption(object):
	def __init__(self, **new):
		self.new = new

	def __enter__(self):
		self.buffer = current.buffer
		self.old = _set_dict(self.buffer.options, self.new, _set_bufoption)[0]

	def __exit__(self, *args):
		self.buffer.options.update(self.old)


class _WithMode(object):
	def __init__(self, new):
		self.new = new

	def __enter__(self):
		self.old = _mode
		_start_mode(self.new)
		return _get_segment_info()

	def __exit__(self, *args):
		_start_mode(self.old)


class _WithDict(object):
	def __init__(self, d, **new):
		self.new = new
		self.d = d

	def __enter__(self):
		self.old, self.na = _set_dict(self.d, self.new)

	def __exit__(self, *args):
		self.d.update(self.old)
		for k in self.na:
			self.d.pop(k)


class _WithSplit(object):
	def __enter__(self):
		_split()

	def __exit__(self, *args):
		_close(2, wipe=False)


class _WithBufName(object):
	def __init__(self, new):
		self.new = new

	def __enter__(self):
		import os
		buffer = current.buffer
		self.buffer = buffer
		self.old = buffer.name
		buffer.name = self.new

	def __exit__(self, *args):
		self.buffer.name = self.old


class _WithNewTabPage(object):
	def __init__(self, *args, **kwargs):
		self.args = args
		self.kwargs = kwargs

	def __enter__(self):
		self.tab = _tabnew(*self.args, **self.kwargs)

	def __exit__(self, *args):
		self.tab._close()


class _WithGlobal(object):
	def __init__(self, **kwargs):
		self.kwargs = kwargs

	def __enter__(self):
		self.empty = object()
		self.old = dict(((key, globals().get(key, self.empty)) for key in self.kwargs))
		globals().update(self.kwargs)

	def __exit__(self, *args):
		for k, v in self.old.items():
			if v is self.empty:
				globals().pop(k, None)
			else:
				globals()[k] = v


@_vim
def _with(key, *args, **kwargs):
	if key == 'buffer':
		return _WithNewBuffer(_edit, *args, **kwargs)
	elif key == 'bufname':
		return _WithBufName(*args, **kwargs)
	elif key == 'mode':
		return _WithMode(*args, **kwargs)
	elif key == 'bufoptions':
		return _WithBufOption(**kwargs)
	elif key == 'options':
		return _WithDict(options, **kwargs)
	elif key == 'globals':
		return _WithDict(vars, **kwargs)
	elif key == 'wvars':
		return _WithDict(current.window.vars, **kwargs)
	elif key == 'environ':
		return _WithDict(_environ, **kwargs)
	elif key == 'split':
		return _WithSplit()
	elif key == 'tabpage':
		return _WithNewTabPage(*args, **kwargs)
	elif key == 'vpos':
		return _WithGlobal(_v_pos=[0, kwargs['line'], kwargs['col'], kwargs['off']])


class error(Exception):
	pass



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import threading
import os

from time import sleep
from itertools import groupby
from signal import SIGKILL
from difflib import ndiff

import pexpect

from powerline.lib.unicode import u

from tests.modules.lib.vterm import VTerm, Dimensions


class MutableDimensions(object):
	def __init__(self, rows, cols):
		super(MutableDimensions, self).__init__()
		self._list = [rows, cols]

	def __getitem__(self, idx):
		return self._list[idx]

	def __setitem__(self, idx, val):
		self._list[idx] = val

	def __iter__(self):
		return iter(self._list)

	def __len__(self):
		return 2

	def __nonzero__(self):
		return True

	__bool__ = __nonzero__

	rows = property(
		fget = lambda self: self._list[0],
		fset = lambda self, val: self._list.__setitem__(0, val),
	)
	cols = property(
		fget = lambda self: self._list[1],
		fset = lambda self, val: self._list.__setitem__(1, val),
	)


class ExpectProcess(threading.Thread):
	def __init__(self, lib, dim, cmd, args, cwd=None, env=None):
		super(ExpectProcess, self).__init__()
		self.vterm = VTerm(lib, dim)
		self.lock = threading.Lock()
		self.dim = Dimensions(*dim)
		self.cmd = cmd
		self.args = args
		self.cwd = cwd
		self.env = env
		self.buffer = []
		self.child_lock = threading.Lock()
		self.shutdown_event = threading.Event()
		self.started_event = threading.Event()

	def run(self):
		with self.child_lock:
			child = pexpect.spawn(self.cmd, self.args, cwd=self.cwd,
			                      env=self.env)
			sleep(0.5)
			child.setwinsize(self.dim.rows, self.dim.cols)
			sleep(0.5)
			self.child = child
			self.started_event.set()
		status = None
		while status is None and not self.shutdown_event.is_set():
			try:
				with self.child_lock:
					s = child.read_nonblocking(size=1024, timeout=0)
					status = child.status
			except pexpect.TIMEOUT:
				pass
			except pexpect.EOF:
				break
			else:
				with self.lock:
					self.vterm.push(s)
					self.buffer.append(s)

		if status is None:
			child.kill(SIGKILL)

	def kill(self):
		self.shutdown_event.set()

	def resize(self, dim):
		with self.child_lock:
			self.dim = Dimensions(*dim)
			self.child.setwinsize(self.dim.rows, self.dim.cols)
			self.vterm.resize(self.dim)

	def __getitem__(self, position):
		with self.lock:
			return self.vterm.vtscreen[position]

	def read(self):
		with self.lock:
			ret = b''.join(self.buffer)
			del self.buffer[:]
			return ret

	def send(self, data):
		with self.child_lock:
			self.child.send(data)

	def get_highlighted_text(self, text, attrs, default_props=(),
	                         use_escapes=False):
		ret = []
		new_attrs = attrs.copy()
		for cell_properties, segment_text in text:
			if use_escapes:
				escapes = ('\033[38;2;{0};{1};{2};48;2;{3};{4};{5}'.format(
					*(cell_properties[0] + cell_properties[1]))) + (
						';1' if cell_properties[2] else ''
					) + (
						';3' if cell_properties[3] else ''
					) + (
						';4' if cell_properties[4] else ''
					) + 'm'
				ret.append(escapes + segment_text + '\033[0m')
			else:
				segment_text = segment_text.translate({'{': '{{', '}': '}}'})
				if cell_properties not in new_attrs:
					new_attrs[cell_properties] = len(new_attrs) + 1
				props_name = new_attrs[cell_properties]
				if props_name in default_props:
					ret.append(segment_text)
				else:
					ret.append('{' + str(props_name) + ':' + segment_text + '}')
		return ''.join(ret), new_attrs

	def get_row(self, row, attrs, default_props=(), use_escapes=False):
		with self.lock:
			return self.get_highlighted_text((
				(key, ''.join((cell.text for cell in subline)))
				for key, subline in groupby((
					self.vterm.vtscreen[row, col]
					for col in range(self.dim.cols)
				), lambda cell: cell.cell_properties_key)
			), attrs, default_props, use_escapes)

	def get_screen(self, attrs, default_props=(), use_escapes=False):
		lines = []
		for row in range(self.dim.rows):
			line, attrs = self.get_row(row, attrs, default_props, use_escapes)
			lines.append(line)
		return '\n'.join(lines), attrs


def test_expected_result(p, test, last_attempt, last_attempt_cb, attempts):
	debugging_tests = not not os.environ.get('_POWERLINE_DEBUGGING_TESTS')
	expected_text, attrs = test['expected_result']
	result = None
	while attempts:
		if 'row' in test:
			row = test['row']
		else:
			row = p.dim.rows - 1
			while row >= 0 and not p[row, 0].text:
				row -= 1
			if row < 0:
				row = 0
		actual_text, all_attrs = p.get_row(row, attrs)
		if actual_text == expected_text:
			return True
		attempts -= 1
		print('Actual result does not match expected for row {0}. Attempts left: {1}.'.format(
			row, attempts))
		sleep(2)
	print('Result (row {0}):'.format(row))
	print(actual_text)
	print('Expected:')
	print(expected_text)
	print('Attributes:')
	for v, k in sorted(
		((v, k) for k, v in all_attrs.items()),
		key=(lambda t: '%02u'.format(t[0]) if isinstance(t[0], int) else t[0]),
	):
		print('{k!r}: {v!r},'.format(v=v, k=k))
	print('Screen:')
	screen, screen_attrs = p.get_screen(attrs, use_escapes=debugging_tests)
	print(screen)
	print(screen_attrs)
	print('_' * 80)
	print('Diff:')
	print('=' * 80)
	print(''.join((
		u(line) for line in ndiff([actual_text + '\n'], [expected_text + '\n']))
	))
	if last_attempt and last_attempt_cb:
		last_attempt_cb()
	return False


ENV_BASE = {
	# Reasoning:
	# 1. vt* TERMs (used to be vt100 here) make tmux-1.9 use different and
	#    identical colors for inactive windows. This is not like tmux-1.6: 
	#    foreground color is different from separator color and equal to (0, 
	#    102, 153) for some reason (separator has correct color). tmux-1.8 is 
	#    fine, so are older versions (though tmux-1.6 and tmux-1.7 do not have 
	#    highlighting for previously active window) and my system tmux-1.9a.
	# 2. screen, xterm and some other non-256color terminals both have the same
	#    issue and make libvterm emit complains like `Unhandled CSI SGR 3231`.
	# 3. screen-256color, xterm-256color and other -256color terminals make
	#    libvterm emit complains about unhandled escapes to stderr.
	# 4. `st-256color` does not have any of the above problems, but it may be
	#    not present on the target system because it is installed with 
	#    x11-terms/st and not with sys-libs/ncurses.
	#
	# For the given reasons decision was made: to fix tmux-1.9 tests and not 
	# make libvterm emit any data to stderr st-256color $TERM should be used, up 
	# until libvterm has its own terminfo database entry (if it ever will). To 
	# make sure that relevant terminfo entry is present on the target system it 
	# should be distributed with powerline test package. To make distribution 
	# not require modifying anything outside of powerline test directory 
	# TERMINFO variable is set.
	#
	# This fix propagates to non-tmux vterm tests just in case.
	'TERM': 'st-256color',
	# Also $TERMINFO definition in get_env

	'POWERLINE_CONFIG_PATHS': os.path.abspath('powerline/config_files'),
	'POWERLINE_COMMAND': 'powerline-render',
	'LD_LIBRARY_PATH': os.environ.get('LD_LIBRARY_PATH', ''),
	'PYTHONPATH': os.environ.get('PYTHONPATH', ''),
}


def get_env(vterm_path, test_dir, *args, **kwargs):
	env = ENV_BASE.copy()
	env.update({
		'TERMINFO': os.path.join(test_dir, 'terminfo'),
		'PATH': vterm_path,
		'SHELL': os.path.join(vterm_path, 'bash'),
	})
	env.update(*args, **kwargs)
	return env


def do_terminal_tests(tests, cmd, dim, args, env, suite, cwd=None, fin_cb=None,
                      last_attempt_cb=None, attempts=None):
	debugging_tests = not not os.environ.get('_POWERLINE_DEBUGGING_TESTS')
	default_attempts = 2 if debugging_tests else 3
	if attempts is None:
		attempts = default_attempts
	lib = os.environ.get('POWERLINE_LIBVTERM')
	if not lib:
		if os.path.exists('tests/bot-ci/deps/libvterm/libvterm.so'):
			lib = 'tests/bot-ci/deps/libvterm/libvterm.so'
		else:
			lib = 'libvterm.so'

	while attempts:
		try:
			p = ExpectProcess(
				lib=lib,
				dim=dim,
				cmd=cmd,
				args=args,
				cwd=cwd,
				env=env,
			)
			p.start()
			p.started_event.wait()

			ret = True

			for i, test in enumerate(tests):
				with suite.test(test.get('name', 'test_{0}'.format(i)),
				                attempts - 1) as ptest:
					try:
						test_prep = test['prep_cb']
					except KeyError:
						pass
					else:
						test_prep(p)
					test_result = test_expected_result(
						p, test, attempts == 0, last_attempt_cb,
						test.get('attempts', default_attempts)
					)
					if not test_result:
						ptest.fail('Result does not match expected')
				ret = ret and test_result

			if ret:
				return ret
		finally:
			if fin_cb:
				fin_cb(p=p, cmd=cmd, env=env)
			p.kill()
			p.join(10)
			assert(not p.isAlive())

		attempts -= 1

	return False



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys

from powerline.lib.monotonic import monotonic
from powerline.lib.inotify import INotifyError
from powerline.lib.path import realpath
from powerline.lib.watcher.inotify import INotifyTreeWatcher, DirTooLarge, NoSuchDir, BaseDirChanged
from powerline.lib.watcher.uv import UvTreeWatcher, UvNotFound


class DummyTreeWatcher(object):
	is_dummy = True

	def __init__(self, basedir):
		self.basedir = realpath(basedir)

	def __call__(self):
		return False


class TreeWatcher(object):
	def __init__(self, pl, watcher_type, expire_time):
		self.watches = {}
		self.last_query_times = {}
		self.expire_time = expire_time * 60
		self.pl = pl
		self.watcher_type = watcher_type

	def get_watcher(self, path, ignore_event):
		if self.watcher_type == 'inotify':
			return INotifyTreeWatcher(path, ignore_event=ignore_event)
		if self.watcher_type == 'uv':
			return UvTreeWatcher(path, ignore_event=ignore_event)
		if self.watcher_type == 'dummy':
			return DummyTreeWatcher(path)
		# FIXME
		if self.watcher_type == 'stat':
			return DummyTreeWatcher(path)
		if self.watcher_type == 'auto':
			if sys.platform.startswith('linux'):
				try:
					return INotifyTreeWatcher(path, ignore_event=ignore_event)
				except (INotifyError, DirTooLarge) as e:
					if not isinstance(e, INotifyError):
						self.pl.warn('Failed to watch path: {0} with error: {1}'.format(path, e))
			try:
				return UvTreeWatcher(path, ignore_event=ignore_event)
			except UvNotFound:
				pass
			return DummyTreeWatcher(path)
		else:
			raise ValueError('Unknown watcher type: {0}'.format(self.watcher_type))

	def watch(self, path, ignore_event=None):
		path = realpath(path)
		w = self.get_watcher(path, ignore_event)
		self.watches[path] = w
		return w

	def expire_old_queries(self):
		pop = []
		now = monotonic()
		for path, lt in self.last_query_times.items():
			if now - lt > self.expire_time:
				pop.append(path)
		for path in pop:
			del self.last_query_times[path]

	def __call__(self, path, ignore_event=None):
		path = realpath(path)
		self.expire_old_queries()
		self.last_query_times[path] = monotonic()
		w = self.watches.get(path, None)
		if w is None:
			try:
				self.watch(path, ignore_event=ignore_event)
			except NoSuchDir:
				pass
			return True
		try:
			return w()
		except BaseDirChanged:
			self.watches.pop(path, None)
			return True
		except DirTooLarge as e:
			self.pl.warn(str(e))
			self.watches[path] = DummyTreeWatcher(path)
			return False



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.theme import requires_segment_info


@requires_segment_info
def prompt_count(pl, segment_info):
	return str(segment_info['ipython'].prompt_count)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os

from powerline.lib.config import ConfigLoader

from tests.modules import TestCase
from tests.modules.lib.fsconfig import FSTree


FILE_ROOT = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'cfglib')


class LoadedList(list):
	def pop_all(self):
		try:
			return self[:]
		finally:
			self[:] = ()


loaded = LoadedList()


def on_load(key):
	loaded.append(key)


def check_file(path):
	if os.path.exists(path):
		return path
	else:
		raise IOError


class TestLoaderCondition(TestCase):
	def test_update_missing(self):
		loader = ConfigLoader(run_once=True)
		fpath = os.path.join(FILE_ROOT, 'file.json')
		self.assertRaises(IOError, loader.load, fpath)
		loader.register_missing(check_file, on_load, fpath)
		loader.update()  # This line must not raise IOError
		with FSTree({'file': {'test': 1}}, root=FILE_ROOT):
			loader.update()
			self.assertEqual(loader.load(fpath), {'test': 1})
			self.assertEqual(loaded.pop_all(), [fpath])


if __name__ == '__main__':
	from tests.modules import main
	main()



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import os

from functools import wraps
from copy import deepcopy

import tests.modules.vim as vim_module

from tests.modules import TestCase
from tests.modules.lib.config_mock import (get_powerline, get_powerline_raw,
                                           swap_attributes, UT)
from tests.modules.lib import Args, replace_item


def highlighted_string(s, group, **kwargs):
	ret = {
		'type': 'string',
		'contents': s,
		'highlight_groups': [group],
	}
	ret.update(kwargs)
	return ret


config = {
	'config': {
		'common': {
			'interval': 0,
			'watcher': 'test',
		},
		'ext': {
			'test': {
				'theme': 'default',
				'colorscheme': 'default',
			},
			'vim': {
				'theme': 'default',
				'colorscheme': 'default',
			},
			'shell': {
				'theme': 'default',
				'colorscheme': 'default',
			},
			'wm': {
				'theme': 'default',
				'colorscheme': 'default',
			},
		},
	},
	'colors': {
		'colors': {
			'col1': 1,
			'col2': 2,
			'col3': 3,
			'col4': 4,
			'col5': 5,
			'col6': 6,
			'col7': 7,
			'col8': 8,
			'col9': 9,
			'col10': 10,
			'col11': 11,
			'col12': 12,
		},
		'gradients': {
		},
	},
	'colorschemes/test/__main__': {
		'groups': {
			'm1': 'g1',
			'm2': 'g3',
			'm3': {'fg': 'col11', 'bg': 'col12', 'attrs': []},
		}
	},
	'colorschemes/default': {
		'groups': {
			'g1': {'fg': 'col5', 'bg': 'col6', 'attrs': []},
			'g2': {'fg': 'col7', 'bg': 'col8', 'attrs': []},
			'g3': {'fg': 'col9', 'bg': 'col10', 'attrs': []},
		}
	},
	'colorschemes/test/default': {
		'groups': {
			'str1': {'fg': 'col1', 'bg': 'col2', 'attrs': ['bold']},
			'str2': {'fg': 'col3', 'bg': 'col4', 'attrs': ['underline']},
			'd1': 'g2',
			'd2': 'm2',
			'd3': 'm3',
		},
	},
	'colorschemes/vim/default': {
		'groups': {
			'environment': {'fg': 'col3', 'bg': 'col4', 'attrs': ['underline']},
		},
	},
	'colorschemes/wm/default': {
		'groups': {
			'hl1': {'fg': 'col1', 'bg': 'col2', 'attrs': ['underline']},
			'hl2': {'fg': 'col2', 'bg': 'col1', 'attrs': []},
			'hl3': {'fg': 'col3', 'bg': 'col1', 'attrs': ['underline']},
			'hl4': {'fg': 'col2', 'bg': 'col4', 'attrs': []},
		},
	},
	'themes/test/default': {
		'segments': {
			'left': [
				highlighted_string('s', 'str1', width='auto'),
				highlighted_string('g', 'str2'),
			],
			'right': [
				highlighted_string('f', 'str2', width='auto', align='r'),
			],
		},
	},
	'themes/' + UT: {
		'dividers': {
			'left': {
				'hard': '>>',
				'soft': '>',
			},
			'right': {
				'hard': '<<',
				'soft': '<',
			},
		},
		'spaces': 0,
	},
	'themes/test/__main__': {
		'dividers': {
			'right': {
				'soft': '|',
			},
		},
	},
	'themes/vim/default': {
		'segments': {
			'left': [
				{
					'function': 'powerline.segments.common.env.environment',
					'args': {
						'variable': 'TEST',
					},
				},
			],
		},
	},
	'themes/shell/default': {
		'default_module': 'powerline.segments.common',
		'segments': {
			'left': [
				highlighted_string('s', 'g1', width='auto'),
			],
		},
	},
	'themes/wm/default': {
		'default_module': 'powerline.segments.common',
		'segments': {
			'left': [
				highlighted_string('A', 'hl1'),
				highlighted_string('B', 'hl2'),
			],
			'right': [
				highlighted_string('C', 'hl3'),
				highlighted_string('D', 'hl4'),
			],
		},
	},
}


def with_new_config(func):
	@wraps(func)
	def f(self):
		return func(self, deepcopy(config))
	return f


def add_args(func):
	@wraps(func)
	def f(self):
		new_config = deepcopy(config)
		with get_powerline(new_config, run_once=True, simpler_renderer=True) as p:
			func(self, p, new_config)
	return f


class TestRender(TestCase):
	def assertRenderEqual(self, p, output, **kwargs):
		self.assertEqual(p.render(**kwargs).replace(' ', ' '), output)

	def assertRenderLinesEqual(self, p, output, **kwargs):
		self.assertEqual([l.replace(' ', ' ') for l in p.render_above_lines(**kwargs)], output)


class TestLines(TestRender):
	@add_args
	def test_without_above(self, p, config):
		self.assertRenderEqual(p, '{121} s{24}>>{344}g{34}>{34}|{344}f {--}')
		self.assertRenderEqual(p, '{121} s {24}>>{344}g{34}>{34}|{344}f {--}', width=10)
		# self.assertRenderEqual(p, '{121} s {24}>>{344}g{34}>{34}|{344} f {--}', width=11)
		self.assertEqual(list(p.render_above_lines()), [])

	@with_new_config
	def test_with_above(self, config):
		old_segments = deepcopy(config['themes/test/default']['segments'])
		config['themes/test/default']['segments']['above'] = [old_segments]
		with get_powerline(config, run_once=True, simpler_renderer=True) as p:
			self.assertRenderLinesEqual(p, [
				'{121} s{24}>>{344}g{34}>{34}|{344}f {--}',
			])
			self.assertRenderLinesEqual(p, [
				'{121} s {24}>>{344}g{34}>{34}|{344}f {--}',
			], width=10)

		config['themes/test/default']['segments']['above'] = [old_segments] * 2
		with get_powerline(config, run_once=True, simpler_renderer=True) as p:
			self.assertRenderLinesEqual(p, [
				'{121} s{24}>>{344}g{34}>{34}|{344}f {--}',
				'{121} s{24}>>{344}g{34}>{34}|{344}f {--}',
			])
			self.assertRenderLinesEqual(p, [
				'{121} s {24}>>{344}g{34}>{34}|{344}f {--}',
				'{121} s {24}>>{344}g{34}>{34}|{344}f {--}',
			], width=10)


class TestSegments(TestRender):
	@add_args
	def test_display(self, p, config):
		config['themes/test/default']['segments']['left'][0]['display'] = False
		config['themes/test/default']['segments']['left'][1]['display'] = True
		config['themes/test/default']['segments']['right'][0]['display'] = False
		self.assertRenderEqual(p, '{344} g{4-}>>{--}')


class TestColorschemesHierarchy(TestRender):
	@add_args
	def test_group_redirects(self, p, config):
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('a', 'd1', draw_hard_divider=False),
				highlighted_string('b', 'd2', draw_hard_divider=False),
				highlighted_string('c', 'd3', draw_hard_divider=False),
				highlighted_string('A', 'm1', draw_hard_divider=False),
				highlighted_string('B', 'm2', draw_hard_divider=False),
				highlighted_string('C', 'm3', draw_hard_divider=False),
				highlighted_string('1', 'g1', draw_hard_divider=False),
				highlighted_string('2', 'g2', draw_hard_divider=False),
				highlighted_string('3', 'g3', draw_hard_divider=False),
			],
			'right': [],
		}
		self.assertRenderEqual(p, '{78} a{910}b{1112}c{56}A{910}B{1112}C{56}1{78}2{910}3{--}')
		self.assertEqual(p.logger._pop_msgs(), [])

	@add_args
	def test_group_redirects_no_main(self, p, config):
		del config['colorschemes/test/__main__']
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('a', 'd1', draw_hard_divider=False),
				highlighted_string('1', 'g1', draw_hard_divider=False),
				highlighted_string('2', 'g2', draw_hard_divider=False),
				highlighted_string('3', 'g3', draw_hard_divider=False),
			],
			'right': [],
		}
		self.assertRenderEqual(p, '{78} a{56}1{78}2{910}3{--}')
		self.assertEqual(p.logger._pop_msgs(), [])

	@add_args
	def test_group_redirects_no_top_default(self, p, config):
		del config['colorschemes/default']
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('c', 'd3', draw_soft_divider=False),
				highlighted_string('C', 'm3', draw_hard_divider=False),
			],
			'right': [],
		}
		self.assertRenderEqual(p, '{1112} c{1112}C{--}')
		self.assertEqual(p.logger._pop_msgs(), [])

	@add_args
	def test_group_redirects_no_test_default(self, p, config):
		del config['colorschemes/test/default']
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('A', 'm1', draw_hard_divider=False),
				highlighted_string('B', 'm2', draw_hard_divider=False),
				highlighted_string('C', 'm3', draw_hard_divider=False),
				highlighted_string('1', 'g1', draw_hard_divider=False),
				highlighted_string('2', 'g2', draw_hard_divider=False),
				highlighted_string('3', 'g3', draw_hard_divider=False),
			],
			'right': [],
		}
		self.assertRenderEqual(p, '{56} A{910}B{1112}C{56}1{78}2{910}3{--}')
		self.assertEqual(p.logger._pop_msgs(), [])

	@add_args
	def test_group_redirects_only_main(self, p, config):
		del config['colorschemes/default']
		del config['colorschemes/test/default']
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('C', 'm3', draw_hard_divider=False),
			],
			'right': [],
		}
		# Powerline is not able to work without default colorscheme 
		# somewhere, thus it will output error string
		self.assertRenderEqual(p, 'colorschemes/test/default')
		self.assertEqual(p.logger._pop_msgs(), [
			'exception:test:powerline:Failed to load colorscheme: colorschemes/default',
			'exception:test:powerline:Failed to load colorscheme: colorschemes/test/default',
			'exception:test:powerline:Failed to create renderer: colorschemes/test/default',
			'exception:test:powerline:Failed to render: colorschemes/test/default',
		])

	@add_args
	def test_group_redirects_only_top_default(self, p, config):
		del config['colorschemes/test/__main__']
		del config['colorschemes/test/default']
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('1', 'g1', draw_hard_divider=False),
				highlighted_string('2', 'g2', draw_hard_divider=False),
				highlighted_string('3', 'g3', draw_hard_divider=False),
			],
			'right': [],
		}
		self.assertRenderEqual(p, '{56} 1{78}2{910}3{--}')
		self.assertEqual(p.logger._pop_msgs(), [])

	@add_args
	def test_group_redirects_only_test_default(self, p, config):
		del config['colorschemes/default']
		del config['colorschemes/test/__main__']
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('s', 'str1', draw_hard_divider=False),
			],
			'right': [],
		}
		self.assertRenderEqual(p, '{121} s{--}')
		self.assertEqual(p.logger._pop_msgs(), [])


class TestThemeHierarchy(TestRender):
	@add_args
	def test_hierarchy(self, p, config):
		self.assertRenderEqual(p, '{121} s{24}>>{344}g{34}>{34}|{344}f {--}')

	@add_args
	def test_no_main(self, p, config):
		del config['themes/test/__main__']
		self.assertRenderEqual(p, '{121} s{24}>>{344}g{34}>{34}<{344}f {--}')
		self.assertEqual(p.logger._pop_msgs(), [])

	@add_args
	def test_no_powerline(self, p, config):
		config['themes/test/__main__']['dividers'] = config['themes/' + UT]['dividers']
		config['themes/test/__main__']['spaces'] = 1
		del config['themes/' + UT]
		self.assertRenderEqual(p, '{121} s {24}>>{344}g {34}>{34}<{344} f {--}')
		self.assertEqual(p.logger._pop_msgs(), [])

	@add_args
	def test_no_default(self, p, config):
		del config['themes/test/default']
		self.assertRenderEqual(p, 'themes/test/default')
		self.assertEqual(p.logger._pop_msgs(), [
			'exception:test:powerline:Failed to load theme: themes/test/default',
			'exception:test:powerline:Failed to create renderer: themes/test/default',
			'exception:test:powerline:Failed to render: themes/test/default',
		])

	@add_args
	def test_only_default(self, p, config):
		config['themes/test/default']['dividers'] = config['themes/' + UT]['dividers']
		config['themes/test/default']['spaces'] = 1
		del config['themes/test/__main__']
		del config['themes/' + UT]
		self.assertRenderEqual(p, '{121} s {24}>>{344}g {34}>{34}<{344} f {--}')

	@add_args
	def test_only_main(self, p, config):
		del config['themes/test/default']
		del config['themes/' + UT]
		self.assertRenderEqual(p, 'themes/test/default')
		self.assertEqual(p.logger._pop_msgs(), [
			'exception:test:powerline:Failed to load theme: themes/' + UT,
			'exception:test:powerline:Failed to load theme: themes/test/default',
			'exception:test:powerline:Failed to create renderer: themes/test/default',
			'exception:test:powerline:Failed to render: themes/test/default',
		])

	@add_args
	def test_only_powerline(self, p, config):
		del config['themes/test/default']
		del config['themes/test/__main__']
		self.assertRenderEqual(p, 'themes/test/default')
		self.assertEqual(p.logger._pop_msgs(), [
			'exception:test:powerline:Failed to load theme: themes/test/__main__',
			'exception:test:powerline:Failed to load theme: themes/test/default',
			'exception:test:powerline:Failed to create renderer: themes/test/default',
			'exception:test:powerline:Failed to render: themes/test/default',
		])

	@add_args
	def test_nothing(self, p, config):
		del config['themes/test/default']
		del config['themes/' + UT]
		del config['themes/test/__main__']
		self.assertRenderEqual(p, 'themes/test/default')
		self.assertEqual(p.logger._pop_msgs(), [
			'exception:test:powerline:Failed to load theme: themes/' + UT,
			'exception:test:powerline:Failed to load theme: themes/test/__main__',
			'exception:test:powerline:Failed to load theme: themes/test/default',
			'exception:test:powerline:Failed to create renderer: themes/test/default',
			'exception:test:powerline:Failed to render: themes/test/default',
		])


class TestDisplayCondition(TestRender):
	@add_args
	def test_include_modes(self, p, config):
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('s1', 'g1', include_modes=['m1']),
				highlighted_string('s2', 'g1', include_modes=['m1', 'm2']),
				highlighted_string('s3', 'g1', include_modes=['m3']),
			]
		}
		self.assertRenderEqual(p, '{--}')
		self.assertRenderEqual(p, '{56} s1{56}>{56}s2{6-}>>{--}', mode='m1')
		self.assertRenderEqual(p, '{56} s2{6-}>>{--}', mode='m2')
		self.assertRenderEqual(p, '{56} s3{6-}>>{--}', mode='m3')

	@add_args
	def test_exclude_modes(self, p, config):
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('s1', 'g1', exclude_modes=['m1']),
				highlighted_string('s2', 'g1', exclude_modes=['m1', 'm2']),
				highlighted_string('s3', 'g1', exclude_modes=['m3']),
			]
		}
		self.assertRenderEqual(p, '{56} s1{56}>{56}s2{56}>{56}s3{6-}>>{--}')
		self.assertRenderEqual(p, '{56} s3{6-}>>{--}', mode='m1')
		self.assertRenderEqual(p, '{56} s1{56}>{56}s3{6-}>>{--}', mode='m2')
		self.assertRenderEqual(p, '{56} s1{56}>{56}s2{6-}>>{--}', mode='m3')

	@add_args
	def test_exinclude_modes(self, p, config):
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('s1', 'g1', exclude_modes=['m1'], include_modes=['m2']),
				highlighted_string('s2', 'g1', exclude_modes=['m1', 'm2'], include_modes=['m3']),
				highlighted_string('s3', 'g1', exclude_modes=['m3'], include_modes=['m3']),
			]
		}
		self.assertRenderEqual(p, '{--}')
		self.assertRenderEqual(p, '{--}', mode='m1')
		self.assertRenderEqual(p, '{56} s1{6-}>>{--}', mode='m2')
		self.assertRenderEqual(p, '{56} s2{6-}>>{--}', mode='m3')

	@add_args
	def test_exinclude_function_nonexistent_module(self, p, config):
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('s1', 'g1', exclude_function='xxx_nonexistent_module.foo'),
				highlighted_string('s2', 'g1', exclude_function='xxx_nonexistent_module.foo', include_function='xxx_nonexistent_module.bar'),
				highlighted_string('s3', 'g1', include_function='xxx_nonexistent_module.bar'),
			]
		}
		self.assertRenderEqual(p, '{56} s1{56}>{56}s2{56}>{56}s3{6-}>>{--}')

	@add_args
	def test_exinclude_function(self, p, config):
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('s1', 'g1', exclude_function='mod.foo'),
				highlighted_string('s2', 'g1', exclude_function='mod.foo', include_function='mod.bar'),
				highlighted_string('s3', 'g1', include_function='mod.bar'),
			]
		}
		launched = set()
		fool = [None]
		barl = [None]

		def foo(*args, **kwargs):
			launched.add('foo')
			self.assertEqual(set(kwargs.keys()), set(('pl', 'segment_info', 'mode')))
			self.assertEqual(args, ())
			return fool[0]

		def bar(*args, **kwargs):
			launched.add('bar')
			self.assertEqual(set(kwargs.keys()), set(('pl', 'segment_info', 'mode')))
			self.assertEqual(args, ())
			return barl[0]

		with replace_item(sys.modules, 'mod', Args(foo=foo, bar=bar)):
			fool[0] = True
			barl[0] = True
			self.assertRenderEqual(p, '{56} s3{6-}>>{--}')
			self.assertEqual(launched, set(('foo', 'bar')))

			fool[0] = False
			barl[0] = True
			self.assertRenderEqual(p, '{56} s1{56}>{56}s2{56}>{56}s3{6-}>>{--}')
			self.assertEqual(launched, set(('foo', 'bar')))

			fool[0] = False
			barl[0] = False
			self.assertRenderEqual(p, '{56} s1{6-}>>{--}')
			self.assertEqual(launched, set(('foo', 'bar')))

			fool[0] = True
			barl[0] = False
			self.assertRenderEqual(p, '{--}')
			self.assertEqual(launched, set(('foo', 'bar')))

	@add_args
	def test_exinclude_modes_override_functions(self, p, config):
		config['themes/test/default']['segments'] = {
			'left': [
				highlighted_string('s1', 'g1', exclude_function='mod.foo', exclude_modes=['m2']),
				highlighted_string('s2', 'g1', exclude_function='mod.foo', include_modes=['m2']),
				highlighted_string('s3', 'g1', include_function='mod.foo', exclude_modes=['m2']),
				highlighted_string('s4', 'g1', include_function='mod.foo', include_modes=['m2']),
			]
		}
		fool = [None]

		def foo(*args, **kwargs):
			return fool[0]

		with replace_item(sys.modules, 'mod', Args(foo=foo)):
			fool[0] = True
			self.assertRenderEqual(p, '{56} s4{6-}>>{--}', mode='m2')
			self.assertRenderEqual(p, '{56} s3{56}>{56}s4{6-}>>{--}', mode='m1')

			fool[0] = False
			self.assertRenderEqual(p, '{56} s2{56}>{56}s4{6-}>>{--}', mode='m2')
			self.assertRenderEqual(p, '{56} s1{6-}>>{--}', mode='m1')


class TestOuterPadding(TestRender):
	@add_args
	def test_outer_padding_left(self, p, config):
		config['themes/' + UT]['outer_padding'] = 5
		self.assertRenderEqual(p, '{121}     s{24}>>{344}g{4-}>>{--}', side='left')

	@add_args
	def test_outer_padding_right(self, p, config):
		config['themes/' + UT]['outer_padding'] = 5
		self.assertRenderEqual(p, '{4-}<<{344}f     {--}', side='right')

	@add_args
	def test_outer_padding_ten(self, p, config):
		config['themes/' + UT]['outer_padding'] = 10
		self.assertRenderEqual(p, '{121}          s  {24}>>{344}g{34}>{34}|{344} f          {--}', width=30)

	@add_args
	def test_outer_padding_zero(self, p, config):
		config['themes/' + UT]['outer_padding'] = 0
		self.assertRenderEqual(p, '{121}s            {24}>>{344}g{34}>{34}|{344}           f{--}', width=30)


class TestSegmentAttributes(TestRender):
	@add_args
	def test_no_attributes(self, p, config):
		def m1(divider=',', **kwargs):
			return divider.join(kwargs.keys()) + divider
		config['themes/test/default']['segments'] = {
			'left': [
				{
					'function': 'bar.m1'
				}
			]
		}
		with replace_item(sys.modules, 'bar', Args(m1=m1)):
			self.assertRenderEqual(p, '{56} pl,{6-}>>{--}')

	@add_args
	def test_segment_datas(self, p, config):
		def m1(divider=',', **kwargs):
			return divider.join(kwargs.keys()) + divider
		m1.powerline_segment_datas = {
			UT: {
				'args': {
					'divider': ';'
				}
			},
			'ascii': {
				'args': {
					'divider': '--'
				}
			}
		}
		config['themes/test/default']['segments'] = {
			'left': [
				{
					'function': 'bar.m1'
				}
			]
		}
		with replace_item(sys.modules, 'bar', Args(m1=m1)):
			self.assertRenderEqual(p, '{56} pl;{6-}>>{--}')

	@add_args
	def test_expand(self, p, config):
		def m1(divider=',', **kwargs):
			return divider.join(kwargs.keys()) + divider

		def expand(pl, amount, segment, **kwargs):
			return ('-' * amount) + segment['contents']

		m1.expand = expand
		config['themes/test/default']['segments'] = {
			'left': [
				{
					'function': 'bar.m1',
					'width': 'auto'
				}
			]
		}
		with replace_item(sys.modules, 'bar', Args(m1=m1)):
			self.assertRenderEqual(p, '{56} ----pl,{6-}>>{--}', width=10)

	@add_args
	def test_truncate(self, p, config):
		def m1(divider=',', **kwargs):
			return divider.join(kwargs.keys()) + divider

		def truncate(pl, amount, segment, **kwargs):
			return segment['contents'][:-amount]

		m1.truncate = truncate
		config['themes/test/default']['segments'] = {
			'left': [
				{
					'function': 'bar.m1'
				}
			]
		}
		with replace_item(sys.modules, 'bar', Args(m1=m1)):
			self.assertRenderEqual(p, '{56} p{6-}>>{--}', width=4)


class TestSegmentData(TestRender):
	@add_args
	def test_segment_data(self, p, config):
		def m1(**kwargs):
			return 'S'

		def m2(**kwargs):
			return 'S'
		sys.modules['bar'] = Args(m1=m1, m2=m2)
		config['themes/' + UT]['segment_data'] = {
			'm1': {
				'before': '1'
			},
			'bar.m2': {
				'before': '2'
			},
			'n': {
				'before': '3'
			},
			'm2': {
				'before': '4'
			},
		}
		config['themes/test/default']['segments'] = {
			'left': [
				{
					'function': 'bar.m1'
				},
				{
					'function': 'bar.m1',
					'name': 'n'
				},
				{
					'function': 'bar.m2',
					'name': 'n'
				},
				{
					'function': 'bar.m2'
				}
			]
		}
		self.assertRenderEqual(p, '{56} 1S{56}>{56}3S{610}>>{910}3S{910}>{910}2S{10-}>>{--}')


class TestShellEscapes(TestCase):
	@with_new_config
	def test_escapes(self, config):
		from powerline.shell import ShellPowerline
		import powerline as powerline_module
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, ShellPowerline, args=Args(config_path=[''])) as powerline:
				self.assertEqual(powerline.render(segment_info={}, side='left'), '\x1b[0;38;5;5;48;5;6m\xa0s\x1b[0;38;5;6;49;22m>>\x1b[0m')

	@with_new_config
	def test_tmux_escapes(self, config):
		from powerline.shell import ShellPowerline
		import powerline as powerline_module
		config['config']['common']['additional_escapes'] = 'tmux'
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, ShellPowerline, args=Args(config_path=[''])) as powerline:
				self.assertEqual(powerline.render(segment_info={}, side='left'), '\x1bPtmux;\x1b\x1b[0;38;5;5;48;5;6m\x1b\\\xa0s\x1bPtmux;\x1b\x1b[0;38;5;6;49;22m\x1b\\>>\x1bPtmux;\x1b\x1b[0m\x1b\\')

	@with_new_config
	def test_screen_escapes(self, config):
		from powerline.shell import ShellPowerline
		import powerline as powerline_module
		config['config']['common']['additional_escapes'] = 'screen'
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, ShellPowerline, args=Args(config_path=[''])) as powerline:
				self.assertEqual(powerline.render(segment_info={}, side='left'), '\x1bP\x1b\x1b[0;38;5;5;48;5;6m\x1b\\\xa0s\x1bP\x1b\x1b[0;38;5;6;49;22m\x1b\\>>\x1bP\x1b\x1b[0m\x1b\\')

	@with_new_config
	def test_fbterm_escapes(self, config):
		from powerline.shell import ShellPowerline
		import powerline as powerline_module
		config['config']['common']['term_escape_style'] = 'fbterm'
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, ShellPowerline, args=Args(config_path=[''])) as powerline:
				self.assertEqual(powerline.render(segment_info={}, side='left'), '\x1b[0m\x1b[1;5}\x1b[2;6}\xa0s\x1b[0m\x1b[1;6}\x1b[49m\x1b[22m>>\x1b[0m')

	@with_new_config
	def test_fbterm_tmux_escapes(self, config):
		from powerline.shell import ShellPowerline
		import powerline as powerline_module
		config['config']['common']['term_escape_style'] = 'fbterm'
		config['config']['common']['additional_escapes'] = 'tmux'
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, ShellPowerline, args=Args(config_path=[''])) as powerline:
				self.assertEqual(powerline.render(segment_info={}, side='left'), '\x1bPtmux;\x1b\x1b[0m\x1b\x1b[1;5}\x1b\x1b[2;6}\x1b\\\xa0s\x1bPtmux;\x1b\x1b[0m\x1b\x1b[1;6}\x1b\x1b[49m\x1b\x1b[22m\x1b\\>>\x1bPtmux;\x1b\x1b[0m\x1b\\')

	@with_new_config
	def test_fbterm_screen_escapes(self, config):
		from powerline.shell import ShellPowerline
		import powerline as powerline_module
		config['config']['common']['term_escape_style'] = 'fbterm'
		config['config']['common']['additional_escapes'] = 'screen'
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, ShellPowerline, args=Args(config_path=[''])) as powerline:
				self.assertEqual(powerline.render(segment_info={}, side='left'), '\x1bP\x1b\x1b[0m\x1b\x1b[1;5}\x1b\x1b[2;6}\x1b\\\xa0s\x1bP\x1b\x1b[0m\x1b\x1b[1;6}\x1b\x1b[49m\x1b\x1b[22m\x1b\\>>\x1bP\x1b\x1b[0m\x1b\\')

	@with_new_config
	def test_term_truecolor_escapes(self, config):
		from powerline.shell import ShellPowerline
		import powerline as powerline_module
		config['config']['common']['term_truecolor'] = True
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, ShellPowerline, args=Args(config_path=[''])) as powerline:
				self.assertEqual(powerline.render(segment_info={}, side='left'), '\x1b[0;38;2;192;0;192;48;2;0;128;128m\xa0s\x1b[0;38;2;0;128;128;49;22m>>\x1b[0m')

	@with_new_config
	def test_term_truecolor_fbterm_escapes(self, config):
		from powerline.shell import ShellPowerline
		import powerline as powerline_module
		config['config']['common']['term_escape_style'] = 'fbterm'
		config['config']['common']['term_truecolor'] = True
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, ShellPowerline, args=Args(config_path=[''])) as powerline:
				self.assertEqual(powerline.render(segment_info={}, side='left'), '\x1b[0m\x1b[1;5}\x1b[2;6}\xa0s\x1b[0m\x1b[1;6}\x1b[49m\x1b[22m>>\x1b[0m')

	@with_new_config
	def test_term_truecolor_tmux_escapes(self, config):
		from powerline.shell import ShellPowerline
		import powerline as powerline_module
		config['config']['common']['term_truecolor'] = True
		config['config']['common']['additional_escapes'] = 'tmux'
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, ShellPowerline, args=Args(config_path=[''])) as powerline:
				self.assertEqual(powerline.render(segment_info={}, side='left'), '\x1bPtmux;\x1b\x1b[0;38;2;192;0;192;48;2;0;128;128m\x1b\\\xa0s\x1bPtmux;\x1b\x1b[0;38;2;0;128;128;49;22m\x1b\\>>\x1bPtmux;\x1b\x1b[0m\x1b\\')

	@with_new_config
	def test_term_truecolor_screen_escapes(self, config):
		from powerline.shell import ShellPowerline
		import powerline as powerline_module
		config['config']['common']['term_truecolor'] = True
		config['config']['common']['additional_escapes'] = 'screen'
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, ShellPowerline, args=Args(config_path=[''])) as powerline:
				self.assertEqual(powerline.render(segment_info={}, side='left'), '\x1bP\x1b\x1b[0;38;2;192;0;192;48;2;0;128;128m\x1b\\\xa0s\x1bP\x1b\x1b[0;38;2;0;128;128;49;22m\x1b\\>>\x1bP\x1b\x1b[0m\x1b\\')


class TestVim(TestCase):
	def test_environ_update(self):
		# Regression test: test that segment obtains environment from vim, not 
		# from os.environ.
		with vim_module._with('globals', powerline_config_paths=['/']):
			from powerline.vim import VimPowerline
			import powerline as powerline_module
			with swap_attributes(config, powerline_module):
				with vim_module._with('environ', TEST='abc'):
					with get_powerline_raw(config, VimPowerline) as powerline:
						window = vim_module.current.window
						window_id = 1
						winnr = window.number
						self.assertEqual(powerline.render(window, window_id, winnr), b'%#Pl_3_8404992_4_192_underline#\xc2\xa0abc%#Pl_4_192_NONE_None_NONE#>>')
						vim_module._environ['TEST'] = 'def'
						self.assertEqual(powerline.render(window, window_id, winnr), b'%#Pl_3_8404992_4_192_underline#\xc2\xa0def%#Pl_4_192_NONE_None_NONE#>>')

	def test_local_themes(self):
		# Regression test: VimPowerline.add_local_theme did not work properly.
		from powerline.vim import VimPowerline
		import powerline as powerline_module
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, VimPowerline, replace_gcp=True) as powerline:
				powerline.add_local_theme('tests.modules.matchers.always_true', {
					'segment_data': {
						'foo': {
							'contents': '“bar”'
						}
					},
					'segments': {
						'left': [
							{
								'type': 'string',
								'name': 'foo',
								'highlight_groups': ['g1']
							}
						]
					}
				})
				window = vim_module.current.window
				window_id = 1
				winnr = window.number
				self.assertEqual(powerline.render(window, window_id, winnr), b'%#Pl_5_12583104_6_32896_NONE#\xc2\xa0\xe2\x80\x9cbar\xe2\x80\x9d%#Pl_6_32896_NONE_None_NONE#>>')

	@classmethod
	def setUpClass(cls):
		sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'vim_sys_path')))

	@classmethod
	def tearDownClass(cls):
		sys.path.pop(0)


class TestLemonbar(TestRender):
	def test_lemonbar(self):
		import powerline as powerline_module
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, powerline_module.Powerline, replace_gcp=True, ext='wm', renderer_module='lemonbar') as powerline:
				self.assertRenderEqual(
					powerline,
					'%{l}%{F#ffc00000}%{B#ff008000}%{+u} A%{F-B--u}%{F#ff008000}%{B#ffc00000}>>%{F-B--u}%{F#ff008000}%{B#ffc00000}B%{F-B--u}%{F#ffc00000}>>%{F-B--u}%{r}%{F#ffc00000}<<%{F-B--u}%{F#ff804000}%{B#ffc00000}%{+u}C%{F-B--u}%{F#ff0000c0}%{B#ffc00000}<<%{F-B--u}%{F#ff008000}%{B#ff0000c0}D %{F-B--u}'
				)

	@with_new_config
	def test_lemonbar_escape(self, config):
		import powerline as powerline_module
		config['themes/wm/default']['segments']['left'] = (
			highlighted_string('%{asd}', 'hl1'),
			highlighted_string('10% %', 'hl2'),
		)
		with swap_attributes(config, powerline_module):
			with get_powerline_raw(config, powerline_module.Powerline, replace_gcp=True, ext='wm', renderer_module='lemonbar') as powerline:
				self.assertRenderEqual(
					powerline,
					'%{l}%{F#ffc00000}%{B#ff008000}%{+u} %%{}{asd}%{F-B--u}%{F#ff008000}%{B#ffc00000}>>%{F-B--u}%{F#ff008000}%{B#ffc00000}10%%{} %%{}%{F-B--u}%{F#ffc00000}>>%{F-B--u}%{r}%{F#ffc00000}<<%{F-B--u}%{F#ff804000}%{B#ffc00000}%{+u}C%{F-B--u}%{F#ff0000c0}%{B#ffc00000}<<%{F-B--u}%{F#ff008000}%{B#ff0000c0}D %{F-B--u}'
				)


if __name__ == '__main__':
	from tests.modules import main
	main()



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.renderers.shell import ShellRenderer


class ZshPromptRenderer(ShellRenderer):
	'''Powerline zsh prompt segment renderer.'''
	escape_hl_start = '%{'
	escape_hl_end = '%}'

	character_translations = ShellRenderer.character_translations.copy()
	character_translations[ord('%')] = '%%'


renderer = ZshPromptRenderer



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os
import json

from subprocess import check_call
from shutil import rmtree
from itertools import chain

from powerline import Powerline


CONFIG_DIR = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'config')


class TestPowerline(Powerline):
	def __init__(self, _paths, *args, **kwargs):
		super(TestPowerline, self).__init__(*args, **kwargs)
		self._paths = _paths

	def get_config_paths(self):
		return self._paths


def mkdir_recursive(directory):
	if os.path.isdir(directory):
		return
	mkdir_recursive(os.path.dirname(directory))
	os.mkdir(directory)


class FSTree(object):
	__slots__ = ('tree', 'p', 'p_kwargs', 'create_p', 'get_config_paths', 'root')

	def __init__(
		self,
		tree,
		p_kwargs={'run_once': True},
		root=CONFIG_DIR,
		get_config_paths=lambda p: (p,),
		create_p=False
	):
		self.tree = tree
		self.root = root
		self.get_config_paths = get_config_paths
		self.create_p = create_p
		self.p = None
		self.p_kwargs = p_kwargs

	def __enter__(self, *args):
		os.mkdir(self.root)
		for k, v in self.tree.items():
			fname = os.path.join(self.root, k) + '.json'
			mkdir_recursive(os.path.dirname(fname))
			with open(fname, 'w') as F:
				json.dump(v, F)
		if self.create_p:
			self.p = TestPowerline(
				_paths=self.get_config_paths(self.root),
				ext='test',
				renderer_module='tests.modules.lib.config_mock',
				**self.p_kwargs
			)
		if os.environ.get('POWERLINE_RUN_LINT_DURING_TESTS'):
			try:
				check_call(chain(['scripts/powerline-lint'], *[
					('-p', d) for d in (
						self.p.get_config_paths() if self.p
						else self.get_config_paths(self.root)
					)
				]))
			except:
				self.__exit__()
				raise
		return self.p and self.p.__enter__(*args)

	def __exit__(self, *args):
		try:
			rmtree(self.root)
		finally:
			if self.p:
				self.p.__exit__(*args)



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import argparse
import codecs
import os
import re

from subprocess import check_output, check_call, CalledProcessError
from getpass import getpass

from github import Github


OVERLAY_NAME = 'raiagent'
OVERLAY = 'leycec/' + OVERLAY_NAME
OVERLAY_BRANCH_FORMAT = 'powerline-release-{0}'


def parse_version(s):
	if s == ('+' * len(s)):
		try:
			last_version = next(iter(sorted([
				tuple(map(int, tag.split('.')))
				for tag in check_output(['git', 'tag', '-l', '[0-9]*.*']).split('\n')[:-1]
			], reverse=True)))
		except StopIteration:
			raise ValueError('No existing versions found')

		version = []
		for i in range(len(s) - 1):
			try:
				version.append(last_version[i])
			except IndexError:
				version.append(0)

		try:
			version.append(last_version[len(s) - 1] + 1)
		except IndexError:
			version.append(1)

		if len(version) == 1:
			version.append(0)

		return tuple(version)
	else:
		return tuple(map(int, s.split('.')))


def setup_py_filter(filter_func):
	with codecs.open('.setup.py.new', 'w', encoding='utf-8') as NS:
		with codecs.open('setup.py', 'r', encoding='utf-8') as OS:
			for line in OS:
				line = filter_func(line)
				NS.write(line)

	os.unlink('setup.py')
	os.rename('.setup.py.new', 'setup.py')


def setup_py_develop_filter(line, version_string):
	if line.startswith('\tbase_version = '):
		line = '\tbase_version = \'' + version_string + '\'\n'
	return line


def setup_py_master_filter(line, version_string):
	if line.startswith('\tversion='):
		line = '\tversion=\'' + version_string + '\',\n'
	elif 'Development Status' in line:
		line = '\t\t\'Development Status :: 5 - Production/Stable\',\n'
	return line


def merge(version_string, rev, **kwargs):
	check_call(['git', 'checkout', rev])

	temp_branch_name = 'release-' + version_string
	check_call(['git', 'checkout', '-b', temp_branch_name])
	setup_py_filter(lambda line: setup_py_develop_filter(line, version_string))
	check_call(['git', 'add', 'setup.py'])
	check_call(['git', 'commit', '-m', 'Update base version'])
	check_call(['git', 'checkout', rev])
	check_call(['git', 'merge', '--no-ff',
	                            '--strategy', 'recursive',
	                            '--strategy-option', 'theirs',
	                            '--commit',
	                            '-m', 'Merge branch \'{0}\' into {1}'.format(temp_branch_name, rev),
	                            temp_branch_name])
	check_call(['git', 'branch', '-d', temp_branch_name])

	rev = check_output(['git', 'rev-parse', 'HEAD']).strip()

	check_call(['git', 'checkout', 'master'])
	try:
		check_call(['git', 'merge', '--no-ff', '--no-commit', '--log', rev])
	except CalledProcessError:
		check_call(['git', 'mergetool', '--tool', 'vimdiff2'])

	setup_py_filter(lambda line: setup_py_master_filter(line, version_string))
	check_call(['git', 'add', 'setup.py'])

	check_call(['git', 'commit'])
	check_call(['git', 'tag', '-m', 'Release ' + version_string, '-a', version_string])


def push(version_string, rev, **kwargs):
	check_call(['git', 'push', 'upstream', 'master'])
	check_call(['git', 'push', 'upstream', version_string])
	check_call(['git', 'push', 'upstream', rev])


def upload(**args):
	check_call(['python', 'setup.py', 'sdist', 'upload'])
	check_call(['python', 'setup.py', 'upload_docs'])


gh = None


def get_gh(user, password):
	global gh
	if not gh:
		gh = Github(user, password)
	return gh


def create_ebuilds(version_string, overlay, user, **kwargs):
	overlay_url = 'git://github.com/' + OVERLAY
	if not os.path.isdir(overlay):
		check_call(['git', 'clone', overlay_url, overlay])
	check_call(['git', 'checkout', 'master'], cwd=overlay)
	check_call(['git', 'pull', '--ff-only', overlay_url, 'master'], cwd=overlay)
	branch = OVERLAY_BRANCH_FORMAT.format(version_string)
	check_call(['git', 'branch', branch], cwd=overlay)
	check_call(['git', 'checkout', branch], cwd=overlay)
	os.environ['DISTDIR'] = '/tmp/powerline-distfiles'
	if not os.path.isdir(os.environ['DISTDIR']):
		os.mkdir(os.environ['DISTDIR'])
	new_files = []
	for category, pn in (
		('app-misc', 'powerline'),
		('app-vim', 'powerline-vim'),
	):
		pdir = os.path.join(overlay, category, pn)
		live_ebuild = None
		for ebuild in os.listdir(pdir):
			if ebuild.endswith('.ebuild') and '9999' in ebuild:
				live_ebuild_base = ebuild
				live_ebuild = os.path.join(pdir, ebuild)
				break
		assert(live_ebuild)
		vcur = os.path.join(pdir, '{0}-{1}.ebuild'.format(pn, version_string))
		if pn == 'powerline-vim':
			with open(live_ebuild) as LEF:
				with open(vcur, 'w') as F:
					dropnext = False
					for line in LEF:
						if line.startswith('EGIT'):
							# Drop all EGIT_… and the next empty line
							dropnext = True
							next_re = re.compile('^$')
							continue
						if dropnext:
							assert(next_re.match(line))
							dropnext = False
							continue
						if line.startswith('# Note the lack of an assignment to ${S}'):
							next_re = re.compile('^#')
							dropnext = True
							line = 'S="${WORKDIR}/${MY_P}"\n'
						if line.startswith('inherit'):
							line = line.replace(' git-r3', '')
							line += '\n'
							line += 'MY_PN="powerline-status"\n'
							line += 'MY_P="${MY_PN}-${PV}"'
							line += '\n'
						elif line.startswith('HOMEPAGE'):
							line += 'SRC_URI="mirror://pypi/p/${MY_PN}/${MY_P}.tar.gz"\n'
						elif line.startswith('KEYWORDS'):
							line = 'KEYWORDS="~amd64 ~ppc ~x86 ~x86-fbsd"\n'
						F.write(line)
		else:
			os.symlink(live_ebuild_base, vcur)
		new_files.append(vcur)
		check_call(['ebuild', vcur, 'manifest'])
		new_files.append(os.path.join(pdir, 'Manifest'))
	check_call(['git', 'add', '--'] + new_files, cwd=overlay)
	check_call(['git', 'commit'] + new_files + ['-m', 'powerline*: Release {0}'.format(version_string)],
	           cwd=overlay)
	check_call(['git', 'push', '-f', 'git@github.com:{0}/{1}'.format(user, OVERLAY_NAME), branch], cwd=overlay)


def update_overlay(version_string, user, password, **kwargs):
	gh = get_gh(user, password)
	overlay = gh.get_repo(OVERLAY)
	overlay.create_pull(
		title='New powerline version: ' + version_string,
		body='Add ebuilds for new powerline version\n\n---\n\nCreated automatically by release script',
		base='master',
		head=user + ':' + OVERLAY_BRANCH_FORMAT.format(version_string),
	)


stages = (
	('merge', merge),
	('push', push),
	('upload', upload),
	('create_ebuilds', create_ebuilds),
	('update_overlay', update_overlay),
)


def create_release(version, user, password=None, run_stages=None, **kwargs):
	version_string = '.'.join((str(v) for v in version))
	if not password:
		password = getpass('Password for {0}: '.format(user))
	for stname, stfunc in stages:
		if run_stages is None or stname in run_stages:
			stfunc(version_string=version_string, user=user, password=password, **kwargs)


p = argparse.ArgumentParser(description='Powerline release script')
p.add_argument('-u', '--user', type=str, metavar='USER', help='Github username.', required=True)
p.add_argument('-v', '--version', type=parse_version, metavar='VERSION', help='Use given version for the release. If version contains only `+\' signs then it will increase latest version number: one `+\' increases major version number (e.g. 1.2.3 -> 2.0), `++\' increases minor version number (e.g. 1.2.3 -> 1.3), `+++\' increases patch level (e.g. 1.2.3 -> 1.2.4). Defaults to `+++\'.', default='+++')
p.add_argument('-r', '--rev', metavar='COMMIT', help='Use given revision for the release. Defaults to `develop\'.', default='develop')
p.add_argument('-s', '--stages', action='append', metavar='STAGE', help='Only run one of the given stages (default to all).', choices=tuple((stname for stname, stfunc in stages)))
p.add_argument('-p', '--password', type=str, metavar='PASS', help='Github password. You will be prompted if it is not supplied.')
p.add_argument('-o', '--overlay', type=str, metavar='PATH', help='Location of the local clone of the {0} overlay. If provided directory does not exist it will be created by “git clone”. Defaults to /tmp/powerline-{0}.'.format(OVERLAY_NAME), default='/tmp/powerline-' + OVERLAY_NAME)


if __name__ == '__main__':
	args = p.parse_args()
	create_release(
		version=args.version,
		rev=args.rev,
		user=args.user,
		password=args.password,
		overlay=args.overlay,
		run_stages=args.stages,
	)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os

from powerline.bindings.vim import vim_getbufoption, buffer_name


def commandt(matcher_info):
	name = buffer_name(matcher_info)
	return (
		vim_getbufoption(matcher_info, 'filetype') == 'command-t'
		or (name and os.path.basename(name) == b'GoToFile')
	)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import json
from collections import namedtuple

from powerline.lib.url import urllib_read, urllib_urlencode
from powerline.lib.threaded import KwThreadedSegment
from powerline.segments import with_docstring


_WeatherKey = namedtuple('Key', 'location_query weather_api_key')


# XXX Warning: module name must not be equal to the segment name as long as this
# segment is imported into powerline.segments.common module.


# Weather condition code descriptions available at
# https://openweathermap.org/weather-conditions
weather_conditions_codes = {
	200: ('stormy',),
	201: ('stormy',),
	202: ('stormy',),
	210: ('stormy',),
	211: ('stormy',),
	212: ('stormy',),
	221: ('stormy',),
	230: ('stormy',),
	231: ('stormy',),
	232: ('stormy',),
	300: ('rainy',),
	301: ('rainy',),
	302: ('rainy',),
	310: ('rainy',),
	311: ('rainy',),
	312: ('rainy',),
	313: ('rainy',),
	314: ('rainy',),
	321: ('rainy',),
	500: ('rainy',),
	501: ('rainy',),
	502: ('rainy',),
	503: ('rainy',),
	504: ('rainy',),
	511: ('snowy',),
	520: ('rainy',),
	521: ('rainy',),
	522: ('rainy',),
	531: ('rainy',),
	600: ('snowy',),
	601: ('snowy',),
	602: ('snowy',),
	611: ('snowy',),
	612: ('snowy',),
	613: ('snowy',),
	615: ('snowy',),
	616: ('snowy',),
	620: ('snowy',),
	621: ('snowy',),
	622: ('snowy',),
	701: ('foggy',),
	711: ('foggy',),
	721: ('foggy',),
	731: ('foggy',),
	741: ('foggy',),
	751: ('foggy',),
	761: ('foggy',),
	762: ('foggy',),
	771: ('foggy',),
	781: ('foggy',),
	800: ('sunny',),
	801: ('cloudy',),
	802: ('cloudy',),
	803: ('cloudy',),
	804: ('cloudy',),
}

weather_conditions_icons = {
	'day':           'DAY',
	'blustery':      'WIND',
	'rainy':         'RAIN',
	'cloudy':        'CLOUDS',
	'snowy':         'SNOW',
	'stormy':        'STORM',
	'foggy':         'FOG',
	'sunny':         'SUN',
	'night':         'NIGHT',
	'windy':         'WINDY',
	'not_available': 'NA',
	'unknown':       'UKN',
}

temp_conversions = {
	'C': lambda temp: temp - 273.15,
	'F': lambda temp: (temp * 9 / 5) - 459.67,
	'K': lambda temp: temp,
}

# Note: there are also unicode characters for units: ℃, ℉ and  K
temp_units = {
	'C': '°C',
	'F': '°F',
	'K': 'K',
}


class WeatherSegment(KwThreadedSegment):
	interval = 600
	default_location = None
	location_urls = {}
	weather_api_key = "fbc9549d91a5e4b26c15be0dbdac3460"

	@staticmethod
	def key(location_query=None, **kwargs):
		try:
			weather_api_key = kwargs["weather_api_key"]
		except KeyError:
			weather_api_key = WeatherSegment.weather_api_key
		return _WeatherKey(location_query, weather_api_key)

	def get_request_url(self, weather_key):
		try:
			return self.location_urls[weather_key]
		except KeyError:
			query_data = {
				"appid": weather_key.weather_api_key
			}
			location_query = weather_key.location_query
			if location_query is None:
				location_data = json.loads(urllib_read('https://freegeoip.app/json/'))
				query_data["lat"] = location_data["latitude"]
				query_data["lon"] = location_data["longitude"]
			else:
				query_data["q"] = location_query
			self.location_urls[location_query] = url = (
				"https://api.openweathermap.org/data/2.5/weather?" +
				urllib_urlencode(query_data))
			return url

	def compute_state(self, weather_key):
		url = self.get_request_url(weather_key)
		raw_response = urllib_read(url)
		if not raw_response:
			self.error('Failed to get response')
			return None

		response = json.loads(raw_response)
		try:
			condition = response['weather'][0]
			condition_code = int(condition['id'])
			temp = float(response['main']['temp'])
		except (KeyError, ValueError):
			self.exception('OpenWeatherMap returned malformed or unexpected response: {0}', repr(raw_response))
			return None

		try:
			icon_names = weather_conditions_codes[condition_code]
		except IndexError:
			icon_names = ('unknown',)
			self.error('Unknown condition code: {0}', condition_code)

		return (temp, icon_names)

	def render_one(self, weather, icons=None, unit='C', temp_format=None, temp_coldest=-30, temp_hottest=40, **kwargs):
		if not weather:
			return None

		temp, icon_names = weather

		for icon_name in icon_names:
			if icons:
				if icon_name in icons:
					icon = icons[icon_name]
					break
		else:
			icon = weather_conditions_icons[icon_names[-1]]

		temp_format = temp_format or ('{temp:.0f}' + temp_units[unit])
		converted_temp = temp_conversions[unit](temp)
		if converted_temp <= temp_coldest:
			gradient_level = 0
		elif converted_temp >= temp_hottest:
			gradient_level = 100
		else:
			gradient_level = (converted_temp - temp_coldest) * 100.0 / (temp_hottest - temp_coldest)
		groups = ['weather_condition_' + icon_name for icon_name in icon_names] + ['weather_conditions', 'weather']
		return [
			{
				'contents': icon + ' ',
				'highlight_groups': groups,
				'divider_highlight_group': 'background:divider',
			},
			{
				'contents': temp_format.format(temp=converted_temp),
				'highlight_groups': ['weather_temp_gradient', 'weather_temp', 'weather'],
				'divider_highlight_group': 'background:divider',
				'gradient_level': gradient_level,
			},
		]


weather = with_docstring(WeatherSegment(),
'''Return weather from OpenWeatherMaps.

Uses GeoIP lookup from https://freegeoip.app to automatically determine
your current location. This should be changed if you’re in a VPN or if your
IP address is registered at another location.

Returns a list of colorized icon and temperature segments depending on
weather conditions.

:param str unit:
	temperature unit, can be one of ``F``, ``C`` or ``K``
:param str location_query:
	location query for your current location, e.g. ``oslo, norway``
:param dict icons:
	dict for overriding default icons, e.g. ``{'heavy_snow' : u'❆'}``
:param str temp_format:
	format string, receives ``temp`` as an argument. Should also hold unit.
:param float temp_coldest:
	coldest temperature. Any temperature below it will have gradient level equal
	to zero.
:param float temp_hottest:
	hottest temperature. Any temperature above it will have gradient level equal
	to 100. Temperatures between ``temp_coldest`` and ``temp_hottest`` receive
	gradient level that indicates relative position in this interval
	(``100 * (cur-coldest) / (hottest-coldest)``).

Divider highlight group used: ``background:divider``.

Highlight groups used: ``weather_conditions`` or ``weather``, ``weather_temp_gradient`` (gradient) or ``weather``.
Also uses ``weather_conditions_{condition}`` for all weather conditions supported by OpenWeatherMap.
''')
